@@ -57,6 +57,7 @@ static gboolean json = FALSE;
 static gboolean compact = FALSE;
 static gboolean headers = FALSE;
 static gboolean raw = FALSE;
+static gboolean ucl_reply = FALSE;
 static gboolean extended_urls = FALSE;
 static gboolean mime_output = FALSE;
 static gboolean empty_input = FALSE;
@@ -130,9 +131,9 @@ static GOptionEntry entries[] =
 	{ "compact", '\0', 0, G_OPTION_ARG_NONE, &compact, "Output compact json reply", NULL},
 	{ "headers", 0, 0, G_OPTION_ARG_NONE, &headers, "Output HTTP headers",
 	  NULL },
-	{ "raw", 0, 0, G_OPTION_ARG_NONE, &raw, "Output raw reply from rspamd",
+	{ "raw", 0, 0, G_OPTION_ARG_NONE, &raw, "Input is a raw file, not an email file",
 	  NULL },
-	{ "ucl", 0, 0, G_OPTION_ARG_NONE, &raw, "Output ucl reply from rspamd",
+	{ "ucl", 0, 0, G_OPTION_ARG_NONE, &ucl_reply, "Output ucl reply from rspamd",
 	  NULL },
 	{ "max-requests", 'n', 0, G_OPTION_ARG_INT, &max_requests,
 	  "Maximum count of parallel requests to rspamd", NULL },
@@ -406,7 +407,7 @@ read_cmd_line (gint *argc, gchar ***argv)
 	}
 
 	if (json || compact) {
-		raw = TRUE;
+		ucl_reply = TRUE;
 	}
 	/* Argc and argv are shifted after this function */
 	g_option_context_free (context);
@@ -635,6 +636,10 @@ add_options (GQueue *opts)
 		ADD_CLIENT_FLAG (flagbuf, "pass_all");
 	}
 
+	if (raw) {
+		ADD_CLIENT_HEADER (opts, "Raw", "yes");
+	}
+
 	if (classifier) {
 		ADD_CLIENT_HEADER (opts, "Classifier", classifier);
 	}
@@ -1035,7 +1040,6 @@ rspamc_counters_output (FILE *out, ucl_object_t *obj)
 	const ucl_object_t *cur, *sym, *weight, *freq, *freq_dev, *nhits;
 	ucl_object_iter_t iter = NULL;
 	gchar fmt_buf[64], dash_buf[82], sym_buf[82];
-	gint l, max_len = INT_MIN, i;
 	static const gint dashes = 44;
 
 	if (obj->type != UCL_ARRAY) {
@@ -1049,16 +1053,17 @@ rspamc_counters_output (FILE *out, ucl_object_t *obj)
 	}
 
 	/* Find maximum width of symbol's name */
+	gint max_len = sizeof("Symbol") - 1;
 	while ((cur = ucl_object_iterate (obj, &iter, true)) != NULL) {
 		sym = ucl_object_lookup (cur, "symbol");
 		if (sym != NULL) {
-			l = sym->len;
-			if (l > max_len) {
-				max_len = MIN (sizeof (dash_buf) - dashes - 1, l);
+			if (sym->len > max_len) {
+				max_len = sym->len;
 			}
 		}
 	}
 
+	max_len = MIN (sizeof (dash_buf) - dashes - 1, max_len);
 	rspamd_snprintf (fmt_buf, sizeof (fmt_buf),
 		"| %%3s | %%%ds | %%7s | %%13s | %%7s |\n", max_len);
 	memset (dash_buf, '-', dashes + max_len);
@@ -1079,7 +1084,7 @@ rspamc_counters_output (FILE *out, ucl_object_t *obj)
 		"| %%3d | %%%ds | %%7.1f | %%6.3f(%%5.3f) | %%7ju |\n", max_len);
 
 	iter = NULL;
-	i = 0;
+	gint i = 0;
 	while ((cur = ucl_object_iterate (obj, &iter, true)) != NULL) {
 		printf (" %s \n", dash_buf);
 		sym = ucl_object_lookup (cur, "symbol");
@@ -1119,21 +1124,23 @@ rspamc_stat_actions (ucl_object_t *obj, GString *out, gint64 scanned)
 	ucl_object_iter_t iter = NULL;
 	gint64 spam, ham;
 
-	if (actions && ucl_object_type (actions) == UCL_OBJECT) {
-		while ((cur = ucl_object_iterate (actions, &iter, true)) != NULL) {
-			gint64 cnt = ucl_object_toint (cur);
-			rspamd_printf_gstring (out, "Messages with action %s: %L"
-				", %.2f%%\n", ucl_object_key (cur), cnt,
-				((gdouble)cnt / (gdouble)scanned) * 100.);
+	if (scanned > 0) {
+		if (actions && ucl_object_type(actions) == UCL_OBJECT) {
+			while ((cur = ucl_object_iterate (actions, &iter, true)) != NULL) {
+				gint64 cnt = ucl_object_toint(cur);
+				rspamd_printf_gstring(out, "Messages with action %s: %L"
+										   ", %.2f%%\n", ucl_object_key(cur), cnt,
+						((gdouble) cnt / (gdouble) scanned) * 100.);
+			}
 		}
-	}
 
-	spam = ucl_object_toint (ucl_object_lookup (obj, "spam_count"));
-	ham = ucl_object_toint (ucl_object_lookup (obj, "ham_count"));
-	rspamd_printf_gstring (out, "Messages treated as spam: %L, %.2f%%\n", spam,
-		((gdouble)spam / (gdouble)scanned) * 100.);
-	rspamd_printf_gstring (out, "Messages treated as ham: %L, %.2f%%\n", ham,
-		((gdouble)ham / (gdouble)scanned) * 100.);
+		spam = ucl_object_toint(ucl_object_lookup(obj, "spam_count"));
+		ham = ucl_object_toint(ucl_object_lookup(obj, "ham_count"));
+		rspamd_printf_gstring(out, "Messages treated as spam: %L, %.2f%%\n", spam,
+				((gdouble) spam / (gdouble) scanned) * 100.);
+		rspamd_printf_gstring(out, "Messages treated as ham: %L, %.2f%%\n", ham,
+				((gdouble) ham / (gdouble) scanned) * 100.);
+	}
 }
 
 static void
@@ -1172,8 +1179,7 @@ static void
 rspamc_stat_output (FILE *out, ucl_object_t *obj)
 {
 	GString *out_str;
-	ucl_object_iter_t iter = NULL;
-	const ucl_object_t *st, *cur;
+	const ucl_object_t *st;
 	gint64 scanned;
 
 	out_str = g_string_sized_new (BUFSIZ);
@@ -1182,9 +1188,7 @@ rspamc_stat_output (FILE *out, ucl_object_t *obj)
 	rspamd_printf_gstring (out_str, "Messages scanned: %L\n",
 		scanned);
 
-	if (scanned > 0) {
-		rspamc_stat_actions (obj, out_str, scanned);
-	}
+	rspamc_stat_actions (obj, out_str, scanned);
 
 	rspamd_printf_gstring (out_str, "Messages learned: %L\n",
 		ucl_object_toint (ucl_object_lookup (obj, "learned")));
@@ -1192,6 +1196,32 @@ rspamc_stat_output (FILE *out, ucl_object_t *obj)
 		ucl_object_toint (ucl_object_lookup (obj, "connections")));
 	rspamd_printf_gstring (out_str, "Control connections count: %L\n",
 		ucl_object_toint (ucl_object_lookup (obj, "control_connections")));
+
+	const ucl_object_t *avg_time_obj = ucl_object_lookup (obj, "scan_times");
+
+	if (avg_time_obj && ucl_object_type (avg_time_obj) == UCL_ARRAY) {
+		ucl_object_iter_t iter = NULL;
+		const ucl_object_t *cur;
+		float sum = 0.0f;
+		volatile float c = 0.0f;
+		unsigned cnt = 0;
+
+		while ((cur = ucl_object_iterate (avg_time_obj, &iter, true)) != NULL) {
+			if (ucl_object_type(cur) == UCL_FLOAT || ucl_object_type(cur) == UCL_INT) {
+				float x = ucl_object_todouble(cur);
+				float y = x - c;
+				float t = sum + y;
+				c = (t - sum) - y;
+				sum = t;
+				cnt ++;
+			}
+		}
+
+		if (cnt > 0) {
+			rspamd_printf_gstring(out_str, "Average scan time: %.3f sec\n", sum / cnt);
+		}
+	}
+
 	/* Pools */
 	rspamd_printf_gstring (out_str, "Pools allocated: %L\n",
 		ucl_object_toint (ucl_object_lookup (obj, "pools_allocated")));
@@ -1228,8 +1258,10 @@ rspamc_stat_output (FILE *out, ucl_object_t *obj)
 
 	st = ucl_object_lookup (obj, "fuzzy_checked");
 	if (st != NULL && ucl_object_type (st) == UCL_ARRAY) {
+		ucl_object_iter_t iter = NULL;
+		const ucl_object_t *cur;
+
 		rspamd_printf_gstring (out_str, "Fuzzy hashes checked: ");
-		iter = NULL;
 
 		while ((cur = ucl_object_iterate (st, &iter, true)) != NULL) {
 			rspamd_printf_gstring (out_str, "%hL ", ucl_object_toint (cur));
@@ -1240,8 +1272,10 @@ rspamc_stat_output (FILE *out, ucl_object_t *obj)
 
 	st = ucl_object_lookup (obj, "fuzzy_found");
 	if (st != NULL && ucl_object_type (st) == UCL_ARRAY) {
+		ucl_object_iter_t iter = NULL;
+		const ucl_object_t *cur;
+
 		rspamd_printf_gstring (out_str, "Fuzzy hashes found: ");
-		iter = NULL;
 
 		while ((cur = ucl_object_iterate (st, &iter, true)) != NULL) {
 			rspamd_printf_gstring (out_str, "%hL ", ucl_object_toint (cur));
@@ -1252,7 +1286,8 @@ rspamc_stat_output (FILE *out, ucl_object_t *obj)
 
 	st = ucl_object_lookup (obj, "statfiles");
 	if (st != NULL && ucl_object_type (st) == UCL_ARRAY) {
-		iter = NULL;
+		ucl_object_iter_t iter = NULL;
+		const ucl_object_t *cur;
 
 		while ((cur = ucl_object_iterate (st, &iter, true)) != NULL) {
 			rspamc_stat_statfile (cur, out_str);
@@ -1383,8 +1418,8 @@ rspamc_mime_output (FILE *out, ucl_object_t *result, GString *input,
 			g_string_erase (symbuf, symbuf->len - 1, 1);
 		}
 
-		folded_symbuf = rspamd_header_value_fold ("X-Spam-Symbols",
-				symbuf->str,
+		folded_symbuf = rspamd_header_value_fold ("X-Spam-Symbols", strlen ("X-Spam-Symbols"),
+				symbuf->str, symbuf->len,
 				0, nl_type, ",");
 		rspamd_printf_gstring (added_headers, "X-Spam-Symbols: %v%s",
 				folded_symbuf, line_end);
@@ -1404,7 +1439,7 @@ rspamc_mime_output (FILE *out, ucl_object_t *result, GString *input,
 			}
 		}
 
-		if (json || raw || compact) {
+		if (json || ucl_reply || compact) {
 			/* We also append json data as a specific header */
 			if (json) {
 				json_header = ucl_object_emit (result,
@@ -1482,7 +1517,7 @@ rspamc_client_execute_cmd (struct rspamc_command *cmd, ucl_object_t *result,
 			rspamc_mime_output (out, result, input, time, err);
 		}
 		else if (result) {
-			if (raw || cmd->command_output_func == NULL) {
+			if (ucl_reply || cmd->command_output_func == NULL) {
 				if (json) {
 					ucl_out = ucl_object_emit (result,
 							compact ? UCL_EMIT_JSON_COMPACT : UCL_EMIT_JSON);
@@ -1554,14 +1589,14 @@ rspamc_client_cb (struct rspamd_client_connection *conn,
 			}
 		}
 		else {
-			if (cmd->need_input) {
+			if (cmd->need_input && !json) {
 				if (!compact) {
 					rspamd_fprintf (out, "Results for file: %s (%.3f seconds)\n",
 							cbdata->filename, diff);
 				}
 			}
 			else {
-				if (!compact) {
+				if (!compact && !json) {
 					rspamd_fprintf (out, "Results for command: %s (%.3f seconds)\n",
 							cmd->name, diff);
 				}
@@ -1571,7 +1606,7 @@ rspamc_client_cb (struct rspamd_client_connection *conn,
 				if (headers && msg != NULL) {
 					rspamc_output_headers (out, msg);
 				}
-				if (raw || cmd->command_output_func == NULL) {
+				if (ucl_reply || cmd->command_output_func == NULL) {
 					if (cmd->need_input) {
 						ucl_object_insert_key (result,
 								ucl_object_fromstring (cbdata->filename),
@@ -1611,15 +1646,15 @@ rspamc_client_cb (struct rspamd_client_connection *conn,
 				rspamd_fprintf (out, "%s\n", err->message);
 
 				if (json && msg != NULL) {
-					const gchar *raw;
+					const gchar *raw_body;
 					gsize rawlen;
 
-					raw = rspamd_http_message_get_body (msg, &rawlen);
+					raw_body = rspamd_http_message_get_body (msg, &rawlen);
 
-					if (raw) {
+					if (raw_body) {
 						/* We can also output the resulting json */
 						rspamd_fprintf (out, "%*s\n", (gint)(rawlen - bodylen),
-								raw);
+								raw_body);
 					}
 				}
 			}
@@ -1727,7 +1762,7 @@ rspamc_process_input (struct ev_loop *ev_base, struct rspamc_command *cmd,
 	else {
 		rspamd_fprintf (stderr, "cannot connect to %s: %s\n", connect_str,
 				strerror (errno));
-		exit (-errno);
+		exit (EXIT_FAILURE);
 	}
 
 	g_free (hostbuf);
