@@ -262,7 +262,7 @@ rspamd_fuzzy_check_ratelimit (struct fuzzy_session *session)
 	}
 	*/
 
-	masked = rspamd_inet_address_copy (session->addr);
+	masked = rspamd_inet_address_copy(session->addr, NULL);
 
 	if (rspamd_inet_address_get_af (masked) == AF_INET) {
 		rspamd_inet_address_apply_mask (masked,
@@ -335,46 +335,39 @@ rspamd_fuzzy_check_ratelimit (struct fuzzy_session *session)
 }
 
 static gboolean
-rspamd_fuzzy_check_client (struct fuzzy_session *session, gboolean is_write)
+rspamd_fuzzy_check_client (struct rspamd_fuzzy_storage_ctx *ctx,
+		rspamd_inet_addr_t *addr)
 {
-	if (session->ctx->blocked_ips != NULL) {
-		if (rspamd_match_radix_map_addr (session->ctx->blocked_ips,
-				session->addr) != NULL) {
+	if (ctx->blocked_ips != NULL) {
+		if (rspamd_match_radix_map_addr (ctx->blocked_ips,
+				addr) != NULL) {
 			return FALSE;
 		}
 	}
 
-	if (is_write) {
-		if (session->ctx->read_only) {
-			return FALSE;
-		}
-
-		if (session->ctx->update_ips != NULL) {
-			if (rspamd_match_radix_map_addr (session->ctx->update_ips,
-					session->addr) == NULL) {
-				return FALSE;
-			}
-			else {
-				return TRUE;
-			}
-		}
-
-		return FALSE;
-	}
-
-	/* Non write */
-	if (session->ctx->ratelimit_buckets) {
-		if (session->ctx->ratelimit_log_only) {
-			(void)rspamd_fuzzy_check_ratelimit (session); /* Check but ignore */
-		}
-		else {
-			return rspamd_fuzzy_check_ratelimit (session);
-		}
-	}
-
 	return TRUE;
 }
 
+static gboolean
+rspamd_fuzzy_check_write (struct fuzzy_session *session)
+{
+	if (session->ctx->read_only) {
+		return FALSE;
+	}
+
+	if (session->ctx->update_ips != NULL) {
+		if (rspamd_match_radix_map_addr (session->ctx->update_ips,
+				session->addr) == NULL) {
+			return FALSE;
+		}
+		else {
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
 static void
 fuzzy_key_stat_dtor (gpointer p)
 {
@@ -465,24 +458,24 @@ rspamd_fuzzy_updates_cb (gboolean success,
 	source = cbdata->source;
 
 	if (success) {
-		rspamd_fuzzy_backend_count (ctx->backend, fuzzy_count_callback, ctx);
+		rspamd_fuzzy_backend_count(ctx->backend, fuzzy_count_callback, ctx);
 
 		msg_info ("successfully updated fuzzy storage %s: %d updates in queue; "
 				  "%d pending currently; "
 				  "%d added; %d deleted; %d extended; %d duplicates",
 				ctx->worker->cf->bind_conf ?
-					 ctx->worker->cf->bind_conf->bind_line :
-					 "unknown",
+				ctx->worker->cf->bind_conf->bind_line :
+				"unknown",
 				cbdata->updates_pending->len,
 				ctx->updates_pending->len,
 				nadded, ndeleted, nextended, nignored);
-		rspamd_fuzzy_backend_version (ctx->backend, source,
+		rspamd_fuzzy_backend_version(ctx->backend, source,
 				fuzzy_update_version_callback, NULL);
 		ctx->updates_failed = 0;
 
 		if (cbdata->final || ctx->worker->state != rspamd_worker_state_running) {
 			/* Plan exit */
-			ev_break (ctx->event_loop, EVBREAK_ALL);
+			ev_break(ctx->event_loop, EVBREAK_ALL);
 		}
 	}
 	else {
@@ -490,36 +483,45 @@ rspamd_fuzzy_updates_cb (gboolean success,
 			msg_err ("cannot commit update transaction to fuzzy backend %s, discard "
 					 "%ud updates after %d retries",
 					ctx->worker->cf->bind_conf ?
-						ctx->worker->cf->bind_conf->bind_line :
-						"unknown",
+					ctx->worker->cf->bind_conf->bind_line :
+					"unknown",
 					cbdata->updates_pending->len,
 					ctx->updates_maxfail);
 			ctx->updates_failed = 0;
 
 			if (cbdata->final || ctx->worker->state != rspamd_worker_state_running) {
 				/* Plan exit */
-				ev_break (ctx->event_loop, EVBREAK_ALL);
+				ev_break(ctx->event_loop, EVBREAK_ALL);
 			}
 		}
 		else {
-			msg_err ("cannot commit update transaction to fuzzy backend %s; "
-					 "%ud updates are still left; %ud currently pending;"
-					 " %d updates left",
-					ctx->worker->cf->bind_conf ?
-						ctx->worker->cf->bind_conf->bind_line :
-						"unknown",
-					cbdata->updates_pending->len,
-					ctx->updates_pending->len,
-					ctx->updates_maxfail - ctx->updates_failed);
-			/* Move the remaining updates to ctx queue */
-			g_array_append_vals (ctx->updates_pending,
-					cbdata->updates_pending->data,
-					cbdata->updates_pending->len);
+			if (ctx->updates_pending) {
+				msg_err ("cannot commit update transaction to fuzzy backend %s; "
+						 "%ud updates are still left; %ud currently pending;"
+						 " %d retries remaining",
+						ctx->worker->cf->bind_conf ?
+							ctx->worker->cf->bind_conf->bind_line : "unknown",
+						cbdata->updates_pending->len,
+						ctx->updates_pending->len,
+						ctx->updates_maxfail - ctx->updates_failed);
+				/* Move the remaining updates to ctx queue */
+				g_array_append_vals(ctx->updates_pending,
+						cbdata->updates_pending->data,
+						cbdata->updates_pending->len);
 
-			if (cbdata->final) {
-				/* Try one more time */
-				rspamd_fuzzy_process_updates_queue (cbdata->ctx, cbdata->source,
-						cbdata->final);
+				if (cbdata->final) {
+					/* Try one more time */
+					rspamd_fuzzy_process_updates_queue(cbdata->ctx, cbdata->source,
+							cbdata->final);
+				}
+			}
+			else {
+				/* We have lost our ctx, so it is a race condition case */
+				msg_err ("cannot commit update transaction to fuzzy backend %s; "
+						 "%ud updates are still left; no more retries: a worker is terminated",
+						ctx->worker->cf->bind_conf ?
+							ctx->worker->cf->bind_conf->bind_line : "unknown",
+						cbdata->updates_pending->len);
 			}
 		}
 	}
@@ -906,10 +908,15 @@ rspamd_fuzzy_check_callback (struct rspamd_fuzzy_reply *result, void *ud)
 	}
 
 	if (!isnan (session->ctx->delay) &&
-			rspamd_get_calendar_ticks () - result->ts < session->ctx->delay &&
 			rspamd_match_radix_map_addr (session->ctx->delay_whitelist,
 					session->addr) == NULL)  {
-		send_flags |= RSPAMD_FUZZY_REPLY_DELAY;
+		gdouble hash_age = rspamd_get_calendar_ticks () - result->ts;
+		gdouble jittered_age = rspamd_time_jitter (session->ctx->delay,
+				session->ctx->delay / 2.0);
+
+		if (hash_age < jittered_age) {
+			send_flags |= RSPAMD_FUZZY_REPLY_DELAY;
+		}
 	}
 
 	/* Refresh hash if found with strong confidence */
@@ -1091,7 +1098,7 @@ rspamd_fuzzy_process_command (struct fuzzy_session *session)
 				session->addr, -1);
 
 		if (ip_stat == NULL) {
-			naddr = rspamd_inet_address_copy (session->addr);
+			naddr = rspamd_inet_address_copy(session->addr, NULL);
 			ip_stat = g_malloc0 (sizeof (*ip_stat));
 			REF_INIT_RETAIN (ip_stat, fuzzy_key_stat_dtor);
 			rspamd_lru_hash_insert (session->key_stat->last_ips,
@@ -1103,7 +1110,18 @@ rspamd_fuzzy_process_command (struct fuzzy_session *session)
 	}
 
 	if (cmd->cmd == FUZZY_CHECK) {
-		if (rspamd_fuzzy_check_client (session, FALSE)) {
+		bool can_continue = true;
+
+		if (session->ctx->ratelimit_buckets) {
+			if (session->ctx->ratelimit_log_only) {
+				(void)rspamd_fuzzy_check_ratelimit (session); /* Check but ignore */
+			}
+			else {
+				can_continue = rspamd_fuzzy_check_ratelimit (session);
+			}
+		}
+
+		if (can_continue) {
 			REF_RETAIN (session);
 			rspamd_fuzzy_backend_check (session->ctx->backend, cmd,
 					rspamd_fuzzy_check_callback, session);
@@ -1122,7 +1140,7 @@ rspamd_fuzzy_process_command (struct fuzzy_session *session)
 		rspamd_fuzzy_make_reply (cmd, &result, session, send_flags);
 	}
 	else {
-		if (rspamd_fuzzy_check_client (session, TRUE)) {
+		if (rspamd_fuzzy_check_write (session)) {
 			/* Check whitelist */
 			if (session->ctx->skip_hashes && cmd->cmd == FUZZY_WRITE) {
 				rspamd_encode_hex_buf (cmd->digest, sizeof (cmd->digest),
@@ -1210,20 +1228,6 @@ rspamd_fuzzy_command_valid (struct rspamd_fuzzy_cmd *cmd, gint r)
 			}
 		}
 		break;
-	case 2:
-		/*
-		 * rspamd 0.8 has slightly different tokenizer then it might be not
-		 * 100% compatible
-		 */
-		if (cmd->shingles_count > 0) {
-			if (r == sizeof (struct rspamd_fuzzy_shingle_cmd)) {
-				ret = RSPAMD_FUZZY_EPOCH8;
-			}
-		}
-		else {
-			ret = RSPAMD_FUZZY_EPOCH8;
-		}
-		break;
 	default:
 		break;
 	}
@@ -1267,7 +1271,8 @@ rspamd_fuzzy_decrypt_command (struct fuzzy_session *s, guchar *buf, gsize buflen
 			RSPAMD_KEYPAIR_KEX, RSPAMD_CRYPTOBOX_MODE_25519);
 
 	if (rk == NULL) {
-		msg_err ("bad key");
+		msg_err ("bad key; ip=%s",
+				rspamd_inet_address_to_string (s->addr));
 		return FALSE;
 	}
 
@@ -1277,7 +1282,8 @@ rspamd_fuzzy_decrypt_command (struct fuzzy_session *s, guchar *buf, gsize buflen
 	if (!rspamd_cryptobox_decrypt_nm_inplace (buf, buflen, hdr.nonce,
 			rspamd_pubkey_get_nm (rk, key->key),
 			hdr.mac, RSPAMD_CRYPTOBOX_MODE_25519)) {
-		msg_err ("decryption failed");
+		msg_err ("decryption failed; ip=%s",
+				rspamd_inet_address_to_string (s->addr));
 		rspamd_pubkey_unref (rk);
 
 		return FALSE;
@@ -1356,7 +1362,7 @@ rspamd_fuzzy_extensions_from_wire (struct fuzzy_session *s, guchar *buf, gsize b
 		p = buf;
 		/*
 		 * Memory layout: n_ext of struct rspamd_fuzzy_cmd_extension
-		 *                payload for each extension in a continious data segment
+		 *                payload for each extension in a continuous data segment
 		 */
 		storage = g_malloc (n_ext * sizeof (struct rspamd_fuzzy_cmd_extension) +
 				st_len);
@@ -1614,14 +1620,24 @@ accept_fuzzy_socket (EV_P_ ev_io *w, int revents)
 #endif
 
 			for (int i = 0; i < r; i ++) {
+				rspamd_inet_addr_t *client_addr;
+
+				client_addr = rspamd_inet_address_from_sa (MSG_FIELD(msg[i], msg_name),
+						MSG_FIELD(msg[i], msg_namelen));
+
+				if (!rspamd_fuzzy_check_client (worker->ctx, client_addr)) {
+					/* Disallow forbidden clients silently */
+					rspamd_inet_address_free (client_addr);
+					continue;
+				}
+
 				session = g_malloc0 (sizeof (*session));
 				REF_INIT_RETAIN (session, fuzzy_session_destroy);
 				session->worker = worker;
 				session->fd = w->fd;
 				session->ctx = worker->ctx;
 				session->time = (guint64) time (NULL);
-				session->addr = rspamd_inet_address_from_sa (MSG_FIELD(msg[i], msg_name),
-						MSG_FIELD(msg[i], msg_namelen));
+				session->addr = client_addr;
 				worker->nconns++;
 
 				/* Each message can have its length in case of recvmmsg */
@@ -1646,7 +1662,7 @@ accept_fuzzy_socket (EV_P_ ev_io *w, int revents)
 						nerrors = g_malloc (sizeof (*nerrors));
 						*nerrors = 1;
 						rspamd_lru_hash_insert (session->ctx->errors_ips,
-								rspamd_inet_address_copy (session->addr),
+								rspamd_inet_address_copy(session->addr, NULL),
 								nerrors, -1, -1);
 					}
 					else {
@@ -1689,6 +1705,7 @@ rspamd_fuzzy_storage_sync (struct rspamd_main *rspamd_main,
 	struct rspamd_control_reply rep;
 
 	rep.reply.fuzzy_sync.status = 0;
+	rep.type = RSPAMD_CONTROL_FUZZY_SYNC;
 
 	if (ctx->backend && worker->index == 0) {
 		rspamd_fuzzy_process_updates_queue (ctx, local_db_name, FALSE);
@@ -1729,8 +1746,8 @@ rspamd_fuzzy_storage_reload (struct rspamd_main *rspamd_main,
 			worker->cf->options, rspamd_main->cfg,
 			&err)) == NULL) {
 		msg_err ("cannot open backend after reload: %e", err);
-		g_error_free (err);
 		rep.reply.reload.status = err->code;
+		g_error_free (err);
 	}
 	else {
 		rep.reply.reload.status = 0;
@@ -1859,7 +1876,7 @@ rspamd_fuzzy_stat_to_ucl (struct rspamd_fuzzy_storage_ctx *ctx, gboolean ip_stat
 	/* Checked by epoch */
 	elt = ucl_object_typed_new (UCL_ARRAY);
 
-	for (i = RSPAMD_FUZZY_EPOCH6; i < RSPAMD_FUZZY_EPOCH_MAX; i++) {
+	for (i = RSPAMD_FUZZY_EPOCH10; i < RSPAMD_FUZZY_EPOCH_MAX; i++) {
 		ucl_array_append (elt,
 				ucl_object_fromint (ctx->stat.fuzzy_hashes_checked[i]));
 	}
@@ -1869,7 +1886,7 @@ rspamd_fuzzy_stat_to_ucl (struct rspamd_fuzzy_storage_ctx *ctx, gboolean ip_stat
 	/* Shingles by epoch */
 	elt = ucl_object_typed_new (UCL_ARRAY);
 
-	for (i = RSPAMD_FUZZY_EPOCH6; i < RSPAMD_FUZZY_EPOCH_MAX; i++) {
+	for (i = RSPAMD_FUZZY_EPOCH10; i < RSPAMD_FUZZY_EPOCH_MAX; i++) {
 		ucl_array_append (elt,
 				ucl_object_fromint (ctx->stat.fuzzy_shingles_checked[i]));
 	}
@@ -1879,7 +1896,7 @@ rspamd_fuzzy_stat_to_ucl (struct rspamd_fuzzy_storage_ctx *ctx, gboolean ip_stat
 	/* Matched by epoch */
 	elt = ucl_object_typed_new (UCL_ARRAY);
 
-	for (i = RSPAMD_FUZZY_EPOCH6; i < RSPAMD_FUZZY_EPOCH_MAX; i++) {
+	for (i = RSPAMD_FUZZY_EPOCH10; i < RSPAMD_FUZZY_EPOCH_MAX; i++) {
 		ucl_array_append (elt,
 				ucl_object_fromint (ctx->stat.fuzzy_hashes_found[i]));
 	}
@@ -1893,10 +1910,16 @@ rspamd_fuzzy_stat_to_ucl (struct rspamd_fuzzy_storage_ctx *ctx, gboolean ip_stat
 static int
 lua_fuzzy_add_pre_handler (lua_State *L)
 {
-	struct rspamd_worker *wrk = *(struct rspamd_worker **)
+	struct rspamd_worker *wrk, **pwrk = (struct rspamd_worker **)
 			rspamd_lua_check_udata (L, 1, "rspamd{worker}");
 	struct rspamd_fuzzy_storage_ctx *ctx;
 
+	if (!pwrk) {
+		return luaL_error (L, "invalid arguments, worker + function are expected");
+	}
+
+	wrk = *pwrk;
+
 	if (wrk && lua_isfunction (L, 2)) {
 		ctx = (struct rspamd_fuzzy_storage_ctx *)wrk->ctx;
 
@@ -1918,10 +1941,16 @@ lua_fuzzy_add_pre_handler (lua_State *L)
 static int
 lua_fuzzy_add_post_handler (lua_State *L)
 {
-	struct rspamd_worker *wrk = *(struct rspamd_worker **)
+	struct rspamd_worker *wrk, **pwrk = (struct rspamd_worker **)
 			rspamd_lua_check_udata (L, 1, "rspamd{worker}");
 	struct rspamd_fuzzy_storage_ctx *ctx;
 
+	if (!pwrk) {
+		return luaL_error (L, "invalid arguments, worker + function are expected");
+	}
+
+	wrk = *pwrk;
+
 	if (wrk && lua_isfunction (L, 2)) {
 		ctx = (struct rspamd_fuzzy_storage_ctx *)wrk->ctx;
 
@@ -2628,6 +2657,7 @@ start_fuzzy (struct rspamd_worker *worker)
 
 	if (worker->index == 0) {
 		g_array_free (ctx->updates_pending, TRUE);
+		ctx->updates_pending = NULL;
 	}
 
 	if (ctx->keypair_cache) {
