@@ -51,9 +51,11 @@ rspamd_archive_dtor (gpointer p)
 	g_ptr_array_free (arch->files, TRUE);
 }
 
-static GString *
+static bool
 rspamd_archive_file_try_utf (struct rspamd_task *task,
-		const gchar *in, gsize inlen)
+							 struct rspamd_archive *arch,
+							 struct rspamd_archive_file *fentry,
+							 const gchar *in, gsize inlen)
 {
 	const gchar *charset = NULL, *p, *end;
 	GString *res;
@@ -74,8 +76,10 @@ rspamd_archive_file_try_utf (struct rspamd_task *task,
 		if (conv == NULL) {
 			msg_info_task ("cannot open converter for %s: %s",
 					charset, u_errorName (uc_err));
+			fentry->flags |= RSPAMD_ARCHIVE_FILE_OBFUSCATED;
+			fentry->fname = g_string_new_len(in, inlen);
 
-			return NULL;
+			return false;
 		}
 
 		tmp = g_malloc (sizeof (*tmp) * (inlen + 1));
@@ -86,9 +90,27 @@ rspamd_archive_file_try_utf (struct rspamd_task *task,
 					charset, u_errorName (uc_err));
 			g_free (tmp);
 
+			fentry->flags |= RSPAMD_ARCHIVE_FILE_OBFUSCATED;
+			fentry->fname = g_string_new_len(in, inlen);
+
 			return NULL;
 		}
 
+		int i = 0;
+
+		while (i < r) {
+			UChar32 uc;
+
+			U16_NEXT(tmp, i, r, uc);
+
+			if (IS_ZERO_WIDTH_SPACE(uc) || u_iscntrl(uc)) {
+				msg_info_task("control character in archive file name found: 0x%02xd "
+							  "(filename=%T)", uc, arch->archive_name);
+				fentry->flags |= RSPAMD_ARCHIVE_FILE_OBFUSCATED;
+				break;
+			}
+		}
+
 		clen = ucnv_getMaxCharSize (utf8_converter);
 		dlen = UCNV_GET_MAX_BYTES_FOR_STRING (r, clen);
 		res = g_string_sized_new (dlen);
@@ -99,6 +121,8 @@ rspamd_archive_file_try_utf (struct rspamd_task *task,
 					charset, u_errorName (uc_err));
 			g_free (tmp);
 			g_string_free (res, TRUE);
+			fentry->flags |= RSPAMD_ARCHIVE_FILE_OBFUSCATED;
+			fentry->fname = g_string_new_len(in, inlen);
 
 			return NULL;
 		}
@@ -108,6 +132,7 @@ rspamd_archive_file_try_utf (struct rspamd_task *task,
 
 		msg_debug_archive ("converted from %s to UTF-8 inlen: %z, outlen: %d",
 				charset, inlen, r);
+		fentry->fname = res;
 	}
 	else {
 		/* Convert unsafe characters to '?' */
@@ -121,13 +146,23 @@ rspamd_archive_file_try_utf (struct rspamd_task *task,
 			}
 			else {
 				g_string_append_c (res, '?');
+
+				if (*p < 0x7f && (g_ascii_iscntrl(*p) || *p == '\0')) {
+					if (!(fentry->flags & RSPAMD_ARCHIVE_FILE_OBFUSCATED)) {
+						msg_info_task("suspicious character in archive file name found: 0x%02xd "
+									  "(filename=%T)",
+									  (int) *p, arch->archive_name);
+						fentry->flags |= RSPAMD_ARCHIVE_FILE_OBFUSCATED;
+					}
+				}
 			}
 
 			p ++;
 		}
+		fentry->fname = res;
 	}
 
-	return res;
+	return true;
 }
 
 static void
@@ -141,7 +176,7 @@ rspamd_archive_process_zip (struct rspamd_task *task,
 	guint32 cd_offset, cd_size, comp_size, uncomp_size, processed = 0;
 	guint16 extra_len, fname_len, comment_len;
 	struct rspamd_archive *arch;
-	struct rspamd_archive_file *f;
+	struct rspamd_archive_file *f = NULL;
 
 	/* Zip files have interesting data at the end of archive */
 	p = part->parsed_data.begin + part->parsed_data.len - 1;
@@ -205,6 +240,9 @@ rspamd_archive_process_zip (struct rspamd_task *task,
 	arch = rspamd_mempool_alloc0 (task->task_pool, sizeof (*arch));
 	arch->files = g_ptr_array_new ();
 	arch->type = RSPAMD_ARCHIVE_ZIP;
+	if (part->cd) {
+		arch->archive_name = &part->cd->filename;
+	}
 	rspamd_mempool_add_destructor (task->task_pool, rspamd_archive_dtor,
 			arch);
 
@@ -239,8 +277,8 @@ rspamd_archive_process_zip (struct rspamd_task *task,
 		}
 
 		f = g_malloc0 (sizeof (*f));
-		f->fname = rspamd_archive_file_try_utf (task,
-				cd + cd_basic_len, fname_len);
+		rspamd_archive_file_try_utf (task, arch, f, cd + cd_basic_len, fname_len);
+
 		f->compressed_size = comp_size;
 		f->uncompressed_size = uncomp_size;
 
@@ -249,11 +287,17 @@ rspamd_archive_process_zip (struct rspamd_task *task,
 		}
 
 		if (f->fname) {
+			if (f->flags & RSPAMD_ARCHIVE_FILE_OBFUSCATED) {
+				arch->flags |= RSPAMD_ARCHIVE_HAS_OBFUSCATED_FILES;
+			}
+
 			g_ptr_array_add (arch->files, f);
 			msg_debug_archive ("found file in zip archive: %v", f->fname);
 		}
 		else {
 			g_free (f);
+
+			return;
 		}
 
 		/* Process extra fields */
@@ -281,10 +325,6 @@ rspamd_archive_process_zip (struct rspamd_task *task,
 	part->part_type = RSPAMD_MIME_PART_ARCHIVE;
 	part->specific.arch = arch;
 
-	if (part->cd) {
-		arch->archive_name = &part->cd->filename;
-	}
-
 	arch->size = part->parsed_data.len;
 }
 
@@ -393,6 +433,9 @@ rspamd_archive_process_rar_v4 (struct rspamd_task *task, const guchar *start,
 	arch = rspamd_mempool_alloc0 (task->task_pool, sizeof (*arch));
 	arch->files = g_ptr_array_new ();
 	arch->type = RSPAMD_ARCHIVE_RAR;
+	if (part->cd) {
+		arch->archive_name = &part->cd->filename;
+	}
 	rspamd_mempool_add_destructor (task->task_pool, rspamd_archive_dtor,
 			arch);
 
@@ -473,19 +516,19 @@ rspamd_archive_process_rar_v4 (struct rspamd_task *task, const guchar *start,
 
 				if (tmp != NULL) {
 					/* Just use ASCII version */
-					f->fname = rspamd_archive_file_try_utf (task, p, tmp - p);
+					rspamd_archive_file_try_utf (task, arch, f, p, tmp - p);
 					msg_debug_archive ("found ascii filename in rarv4 archive: %v",
 							f->fname);
 				}
 				else {
 					/* We have UTF8 filename, use it as is */
-					f->fname = rspamd_archive_file_try_utf (task, p, fname_len);
+					rspamd_archive_file_try_utf (task, arch, f, p, fname_len);
 					msg_debug_archive ("found utf filename in rarv4 archive: %v",
 							f->fname);
 				}
 			}
 			else {
-				f->fname = rspamd_archive_file_try_utf (task, p, fname_len);
+				rspamd_archive_file_try_utf (task, arch, f, p, fname_len);
 				msg_debug_archive ("found ascii (old) filename in rarv4 archive: %v",
 						f->fname);
 			}
@@ -498,6 +541,9 @@ rspamd_archive_process_rar_v4 (struct rspamd_task *task, const guchar *start,
 			}
 
 			if (f->fname) {
+				if (f->flags & RSPAMD_ARCHIVE_FILE_OBFUSCATED) {
+					arch->flags |= RSPAMD_ARCHIVE_HAS_OBFUSCATED_FILES;
+				}
 				g_ptr_array_add (arch->files, f);
 			}
 			else {
@@ -512,7 +558,6 @@ rspamd_archive_process_rar_v4 (struct rspamd_task *task, const guchar *start,
 end:
 	part->part_type = RSPAMD_MIME_PART_ARCHIVE;
 	part->specific.arch = arch;
-	arch->archive_name = &part->cd->filename;
 	arch->size = part->parsed_data.len;
 }
 
@@ -559,6 +604,9 @@ rspamd_archive_process_rar (struct rspamd_task *task,
 	arch = rspamd_mempool_alloc0 (task->task_pool, sizeof (*arch));
 	arch->files = g_ptr_array_new ();
 	arch->type = RSPAMD_ARCHIVE_RAR;
+	if (part->cd) {
+		arch->archive_name = &part->cd->filename;
+	}
 	rspamd_mempool_add_destructor (task->task_pool, rspamd_archive_dtor,
 			arch);
 
@@ -683,11 +731,14 @@ rspamd_archive_process_rar (struct rspamd_task *task,
 			f = g_malloc0 (sizeof (*f));
 			f->uncompressed_size = uncomp_sz;
 			f->compressed_size = comp_sz;
-			f->fname = rspamd_archive_file_try_utf (task, p, fname_len);
+			rspamd_archive_file_try_utf (task, arch, f, p, fname_len);
 
 			if (f->fname) {
 				msg_debug_archive ("added rarv5 file: %v", f->fname);
 				g_ptr_array_add (arch->files, f);
+				if (f->flags & RSPAMD_ARCHIVE_FILE_OBFUSCATED) {
+					arch->flags |= RSPAMD_ARCHIVE_HAS_OBFUSCATED_FILES;
+				}
 			}
 			else {
 				g_free (f);
@@ -736,9 +787,6 @@ rspamd_archive_process_rar (struct rspamd_task *task,
 end:
 	part->part_type = RSPAMD_MIME_PART_ARCHIVE;
 	part->specific.arch = arch;
-	if (part->cd != NULL) {
-		arch->archive_name = &part->cd->filename;
-	}
 	arch->size = part->parsed_data.len;
 }
 
@@ -1261,6 +1309,7 @@ rspamd_7zip_read_substreams_info (struct rspamd_task *task,
 	}
 
 	folder_nstreams = g_alloca (sizeof (guint64) * num_folders);
+	memset (folder_nstreams, 0, sizeof (guint64) * num_folders);
 
 	while (p != NULL && p < end) {
 		/*
@@ -1407,10 +1456,10 @@ rspamd_7zip_read_archive_props (struct rspamd_task *task,
 	 * }
 	 */
 
-	proptype = *p;
-	SZ_SKIP_BYTES(1);
-
 	if (p != NULL) {
+		proptype = *p;
+		SZ_SKIP_BYTES(1);
+
 		while (proptype != 0) {
 			SZ_READ_VINT(proplen);
 
@@ -1703,6 +1752,9 @@ rspamd_archive_process_gzip (struct rspamd_task *task,
 	arch = rspamd_mempool_alloc0 (task->task_pool, sizeof (*arch));
 	arch->files = g_ptr_array_sized_new (1);
 	arch->type = RSPAMD_ARCHIVE_GZIP;
+	if (part->cd) {
+		arch->archive_name = &part->cd->filename;
+	}
 	rspamd_mempool_add_destructor (task->task_pool, rspamd_archive_dtor,
 			arch);
 
@@ -1747,11 +1799,16 @@ rspamd_archive_process_gzip (struct rspamd_task *task,
 					struct rspamd_archive_file *f;
 
 					f = g_malloc0 (sizeof (*f));
-					f->fname = rspamd_archive_file_try_utf (task, fname_start,
-							p - fname_start);
+
+					rspamd_archive_file_try_utf (task, arch, f,
+							fname_start, p - fname_start);
 
 					if (f->fname) {
 						g_ptr_array_add (arch->files, f);
+
+						if (f->flags & RSPAMD_ARCHIVE_FILE_OBFUSCATED) {
+							arch->flags |= RSPAMD_ARCHIVE_HAS_OBFUSCATED_FILES;
+						}
 					}
 					else {
 						/* Invalid filename, skip */
@@ -1832,11 +1889,6 @@ set:
 	/* Set archive data */
 	part->part_type = RSPAMD_MIME_PART_ARCHIVE;
 	part->specific.arch = arch;
-
-	if (part->cd) {
-		arch->archive_name = &part->cd->filename;
-	}
-
 	arch->size = part->parsed_data.len;
 }
 
@@ -1951,7 +2003,10 @@ rspamd_archives_process (struct rspamd_task *task)
 					msg_info_task ("found %s archive with incorrect content-type: %T/%T",
 							rspamd_archive_type_str (arch->type),
 							&part->ct->type, &part->ct->subtype);
-					part->ct->flags |= RSPAMD_CONTENT_TYPE_BROKEN;
+
+					if (!(part->ct->flags & RSPAMD_CONTENT_TYPE_MISSING)) {
+						part->ct->flags |= RSPAMD_CONTENT_TYPE_BROKEN;
+					}
 				}
 			}
 		}
