@@ -20,6 +20,7 @@
 #include "libutil/mem_pool.h"
 #include "libutil/addr.h"
 #include "khash.h"
+#include "contrib/libucl/ucl.h"
 
 #ifdef  __cplusplus
 extern "C" {
@@ -48,6 +49,10 @@ enum rspamd_mime_header_flags {
 	RSPAMD_HEADER_UNIQUE = 1u << 12u,
 	RSPAMD_HEADER_EMPTY_SEPARATOR = 1u << 13u,
 	RSPAMD_HEADER_TAB_SEPARATED = 1u << 14u,
+	RSPAMD_HEADER_MODIFIED = 1u << 15u, /* Means we need to check modified chain */
+	RSPAMD_HEADER_ADDED = 1u << 16u, /* A header has been artificially added */
+	RSPAMD_HEADER_REMOVED = 1u << 17u, /* A header has been artificially removed */
+	RSPAMD_HEADER_NON_EXISTING = 1u << 18u, /* Header was not in the original message */
 };
 
 struct rspamd_mime_header {
@@ -60,58 +65,13 @@ struct rspamd_mime_header {
 	gchar *value;
 	gchar *separator;
 	gchar *decoded;
+	struct rspamd_mime_header *modified_chain; /* Headers modified during transform */
 	struct rspamd_mime_header *prev, *next; /* Headers with the same name */
 	struct rspamd_mime_header *ord_next; /* Overall order of headers, slist */
 };
 
 struct rspamd_mime_headers_table;
 
-enum rspamd_received_type {
-	RSPAMD_RECEIVED_SMTP = 1u << 0u,
-	RSPAMD_RECEIVED_ESMTP = 1u << 1u,
-	RSPAMD_RECEIVED_ESMTPA = 1u << 2u,
-	RSPAMD_RECEIVED_ESMTPS = 1u << 3u,
-	RSPAMD_RECEIVED_ESMTPSA = 1u << 4u,
-	RSPAMD_RECEIVED_LMTP = 1u << 5u,
-	RSPAMD_RECEIVED_IMAP = 1u << 6u,
-	RSPAMD_RECEIVED_LOCAL = 1u << 7u,
-	RSPAMD_RECEIVED_HTTP = 1u << 8u,
-	RSPAMD_RECEIVED_MAPI = 1u << 9u,
-	RSPAMD_RECEIVED_UNKNOWN = 1u << 10u,
-	RSPAMD_RECEIVED_FLAG_ARTIFICIAL =  (1u << 11u),
-	RSPAMD_RECEIVED_FLAG_SSL =  (1u << 12u),
-	RSPAMD_RECEIVED_FLAG_AUTHENTICATED =  (1u << 13u),
-};
-
-#define RSPAMD_RECEIVED_FLAG_TYPE_MASK (RSPAMD_RECEIVED_SMTP| \
-			RSPAMD_RECEIVED_ESMTP| \
-			RSPAMD_RECEIVED_ESMTPA| \
-			RSPAMD_RECEIVED_ESMTPS| \
-			RSPAMD_RECEIVED_ESMTPSA| \
-			RSPAMD_RECEIVED_LMTP| \
-			RSPAMD_RECEIVED_IMAP| \
-			RSPAMD_RECEIVED_LOCAL| \
-			RSPAMD_RECEIVED_HTTP| \
-			RSPAMD_RECEIVED_MAPI| \
-			RSPAMD_RECEIVED_UNKNOWN)
-
-struct rspamd_email_address;
-
-struct rspamd_received_header {
-	const gchar *from_hostname;
-	const gchar *from_ip;
-	const gchar *real_hostname;
-	const gchar *real_ip;
-	const gchar *by_hostname;
-	const gchar *for_mbox;
-	struct rspamd_email_address *for_addr;
-	rspamd_inet_addr_t *addr;
-	struct rspamd_mime_header *hdr;
-	time_t timestamp;
-	gint flags; /* See enum rspamd_received_type */
-	struct rspamd_received_header *prev, *next;
-};
-
 /**
  * Process headers and store them in `target`
  * @param task
@@ -159,7 +119,8 @@ gchar *rspamd_mime_message_id_generate (const gchar *fqdn);
  */
 struct rspamd_mime_header *
 rspamd_message_get_header_array (struct rspamd_task *task,
-								 const gchar *field);
+								const gchar *field,
+								gboolean need_modified);
 
 /**
  * Get an array of header's values with specified header's name using raw headers
@@ -169,7 +130,21 @@ rspamd_message_get_header_array (struct rspamd_task *task,
  */
 struct rspamd_mime_header *
 rspamd_message_get_header_from_hash (struct rspamd_mime_headers_table *hdrs,
-									 const gchar *field);
+									 const gchar *field,
+									 gboolean need_modified);
+
+/**
+ * Modifies a header (or insert one if not found)
+ * @param hdrs
+ * @param hdr_name
+ * @param obj an array of modified values
+ *
+ */
+void
+rspamd_message_set_modified_header (struct rspamd_task *task,
+									struct rspamd_mime_headers_table *hdrs,
+									const gchar *hdr_name,
+									const ucl_object_t *obj);
 
 /**
  * Cleans up hash table of the headers
@@ -185,6 +160,31 @@ struct rspamd_mime_headers_table * rspamd_message_headers_ref (struct rspamd_mim
  */
 struct rspamd_mime_headers_table* rspamd_message_headers_new (void);
 
+/**
+ * Returns size for a headers table
+ * @param hdrs
+ * @return
+ */
+gsize rspamd_mime_headers_count (struct rspamd_mime_headers_table *hdrs);
+
+typedef bool(rspamd_hdr_traverse_func_t)(const gchar *, const struct rspamd_mime_header *, void *);
+/**
+ * Traverse all headers in a table
+ * @param func
+ * @param ud
+ * @return
+ */
+bool rspamd_mime_headers_foreach(const struct rspamd_mime_headers_table *,
+		rspamd_hdr_traverse_func_t func, void *ud);
+
+/**
+ * Strip rfc822 CFWS sequences from a string in place
+ * @param input input
+ * @param len length of the input
+ * @return new length of the input
+ */
+gsize rspamd_strip_smtp_comments_inplace (gchar *input, gsize len);
+
 #ifdef  __cplusplus
 }
 #endif
