@@ -76,6 +76,7 @@ enum rspamd_log_cfg_flags {
 	RSPAMD_LOG_FLAG_USEC = (1 << 3),
 	RSPAMD_LOG_FLAG_RSPAMADM = (1 << 4),
 	RSPAMD_LOG_FLAG_ENFORCED = (1 << 5),
+	RSPAMD_LOG_FLAG_SEVERITY = (1 << 6),
 };
 
 struct rspamd_worker_log_pipe {
@@ -128,7 +129,7 @@ enum rspamd_symbol_flags {
 	RSPAMD_SYMBOL_FLAG_NORMAL = 0,
 	RSPAMD_SYMBOL_FLAG_IGNORE_METRIC = (1 << 1),
 	RSPAMD_SYMBOL_FLAG_ONEPARAM = (1 << 2),
-	RSPAMD_SYMBOL_FLAG_UNGROUPPED = (1 << 3),
+	RSPAMD_SYMBOL_FLAG_UNGROUPED = (1 << 3),
 	RSPAMD_SYMBOL_FLAG_DISABLED = (1 << 4),
 	RSPAMD_SYMBOL_FLAG_UNSCORED = (1 << 5),
 };
@@ -191,6 +192,7 @@ struct rspamd_classifier_config {
 	const gchar *backend;                           /**< name of statfile's backend							*/
 	ucl_object_t *opts;                             /**< other options                                      */
 	GList *learn_conditions;                        /**< list of learn condition callbacks					*/
+	GList *classify_conditions;                     /**< list of classify condition callbacks					*/
 	gchar *name;                                    /**< unique name of classifier							*/
 	guint32 min_tokens;                             /**< minimal number of tokens to process classifier 	*/
 	guint32 max_tokens;                             /**< maximum number of tokens							*/
@@ -312,6 +314,7 @@ struct rspamd_action;
 struct rspamd_config_cfg_lua_script {
 	gint cbref;
 	gint priority;
+	gchar *lua_src_pos;
 	struct rspamd_config_cfg_lua_script *prev, *next;
 };
 
@@ -360,7 +363,6 @@ struct rspamd_config {
 	GHashTable *groups;                            /**< groups of symbols								*/
 	struct rspamd_action *actions;                  /**< all actions of the metric						*/
 
-	gboolean raw_mode;                              /**< work in raw mode instead of utf one				*/
 	gboolean one_shot_mode;                         /**< rules add only one symbol							*/
 	gboolean check_text_attachements;               /**< check text attachements as text					*/
 	gboolean check_all_filters;                     /**< check all filters									*/
@@ -376,6 +378,7 @@ struct rspamd_config {
 	gboolean soft_reject_on_timeout;                /**< If true emit soft reject on task timeout (if not reject) */
 	gboolean public_groups_only;                    /**< Output merely public groups everywhere				*/
 	gboolean enable_test_patterns;                  /**< Enable test patterns								*/
+	gboolean enable_css_parser;                     /**< Enable css parsing in HTML							*/
 
 	gsize max_cores_size;                           /**< maximum size occupied by rspamd core files			*/
 	gsize max_cores_count;                          /**< maximum number of core files						*/
@@ -416,7 +419,7 @@ struct rspamd_config {
 	ucl_object_t *config_comments;                  /**< comments saved from the config						*/
 	ucl_object_t *doc_strings;                      /**< documentation strings for config options			*/
 	GPtrArray *c_modules;                           /**< list of C modules			*/
-	GHashTable *composite_symbols;                 /**< hash of composite symbols indexed by its name		*/
+	void *composites_manager;                       /**< hash of composite symbols indexed by its name		*/
 	GList *classifiers;                             /**< list of all classifiers defined                    */
 	GList *statfiles;                               /**< list of all statfiles in config file order         */
 	GHashTable *classifiers_symbols;                /**< hashtable indexed by symbol name of classifiers    */
@@ -476,14 +479,14 @@ struct rspamd_config {
 	guint max_blas_threads;                         /**< maximum threads for openblas when learning ANN		*/
 	guint max_opts_len;                             /**< maximum length for all options for a symbol		*/
 
-	GList *classify_headers;                        /**< list of headers using for statistics				*/
 	struct module_s **compiled_modules;                /**< list of compiled C modules							*/
-	struct worker_s **compiled_workers;                /**< list of compiled C modules							*/struct rspamd_log_format *log_format;            /**< parsed log format									*/
+	struct worker_s **compiled_workers;                /**< list of compiled C modules							*/
+	struct rspamd_log_format *log_format;            /**< parsed log format									*/
 	gchar *log_format_str;                            /**< raw log format string								*/
 
 	struct rspamd_external_libs_ctx *libs_ctx;        /**< context for external libraries						*/
 	struct rspamd_monitored_ctx *monitored_ctx;        /**< context for monitored resources					*/
-	struct rspamd_redis_pool *redis_pool;            /**< redis connectiosn pool								*/
+	struct rspamd_redis_pool *redis_pool;            /**< redis connection pool								*/
 
 	struct rspamd_re_cache *re_cache;                /**< static regexp cache								*/
 
@@ -580,25 +583,6 @@ enum rspamd_post_load_options {
 gboolean rspamd_config_post_load (struct rspamd_config *cfg,
 								  enum rspamd_post_load_options opts);
 
-/**
- * Calculate checksum for config file
- * @param cfg config file
- */
-gboolean rspamd_config_calculate_checksum (struct rspamd_config *cfg);
-
-
-/**
- * Replace all \" with a single " in given string
- * @param line input string
- */
-void rspamd_config_unescape_quotes (gchar *line);
-
-/*
- * Convert comma separated string to a list of strings
- */
-GList *rspamd_config_parse_comma_list (rspamd_mempool_t *pool,
-									   const gchar *line);
-
 /*
  * Return a new classifier_config structure, setting default and non-conflicting attributes
  */
@@ -857,26 +841,27 @@ gboolean rspamd_ip_is_local_cfg (struct rspamd_config *cfg,
 gboolean rspamd_config_libs (struct rspamd_external_libs_ctx *ctx,
 							 struct rspamd_config *cfg);
 
+
 #define msg_err_config(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         cfg->cfg_pool->tag.tagname, cfg->checksum, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_err_config_forced(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL|RSPAMD_LOG_FORCED, \
         cfg->cfg_pool->tag.tagname, cfg->checksum, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_warn_config(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
         cfg->cfg_pool->tag.tagname, cfg->checksum, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 #define msg_info_config(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
         cfg->cfg_pool->tag.tagname, cfg->checksum, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 extern guint rspamd_config_log_id;
 #define msg_debug_config(...)  rspamd_conditional_debug_fast (NULL, NULL, \
         rspamd_config_log_id, "config", cfg->checksum, \
-        G_STRFUNC, \
+        RSPAMD_LOG_FUNC, \
         __VA_ARGS__)
 
 #ifdef  __cplusplus
