@@ -98,14 +98,14 @@ struct rspamd_regexp_map_helper {
 	while (g_ascii_isspace (*c) && p > c) { c ++; } \
 	key = g_malloc (p - c + 1); \
 	rspamd_strlcpy (key, c, p - c + 1); \
-	key = g_strstrip (key); \
+	stripped_key = g_strstrip (key); \
 } while (0)
 
 #define MAP_STORE_VALUE do { \
 	while (g_ascii_isspace (*c) && p > c) { c ++; } \
 	value = g_malloc (p - c + 1); \
 	rspamd_strlcpy (value, c, p - c + 1); \
-	value = g_strstrip (value); \
+	stripped_value = g_strstrip (value); \
 } while (0)
 
 gchar *
@@ -132,7 +132,7 @@ rspamd_parse_kv_list (
 		map_read_eol,
 	};
 
-	gchar *c, *p, *key = NULL, *value = NULL, *end;
+	gchar *c, *p, *key = NULL, *value = NULL, *stripped_key, *stripped_value, *end;
 	struct rspamd_map *map = data->map;
 	guint line_number = 0;
 
@@ -171,9 +171,9 @@ rspamd_parse_kv_list (
 				if (p - c > 0) {
 					/* Store a single key */
 					MAP_STORE_KEY;
-					func (data->cur_data, key, default_value);
+					func (data->cur_data, stripped_key, default_value);
 					msg_debug_map ("insert key only pair: %s -> %s; line: %d",
-							key, default_value, line_number);
+							stripped_key, default_value, line_number);
 					g_free (key);
 				}
 
@@ -184,9 +184,9 @@ rspamd_parse_kv_list (
 				if (p - c > 0) {
 					/* Store a single key */
 					MAP_STORE_KEY;
-					func (data->cur_data, key, default_value);
+					func (data->cur_data, stripped_key, default_value);
 					msg_debug_map ("insert key only pair: %s -> %s; line: %d",
-							key, default_value, line_number);
+							stripped_key, default_value, line_number);
 					g_free (key);
 				}
 
@@ -254,9 +254,9 @@ rspamd_parse_kv_list (
 				if (p - c > 0) {
 					/* Store a single key */
 					MAP_STORE_KEY;
-					func (data->cur_data, key, default_value);
+					func (data->cur_data, stripped_key, default_value);
 					msg_debug_map ("insert key only pair: %s -> %s; line: %d",
-							key, default_value, line_number);
+							stripped_key, default_value, line_number);
 					g_free (key);
 					key = NULL;
 				}
@@ -267,10 +267,10 @@ rspamd_parse_kv_list (
 				if (p - c > 0) {
 					/* Store a single key */
 					MAP_STORE_KEY;
-					func (data->cur_data, key, default_value);
+					func (data->cur_data, stripped_key, default_value);
 
 					msg_debug_map ("insert key only pair: %s -> %s; line: %d",
-							key, default_value, line_number);
+							stripped_key, default_value, line_number);
 					g_free (key);
 					key = NULL;
 				}
@@ -320,17 +320,17 @@ rspamd_parse_kv_list (
 					if (p - c > 0) {
 						/* Store a single key */
 						MAP_STORE_VALUE;
-						func (data->cur_data, key, value);
+						func (data->cur_data, stripped_key, stripped_value);
 						msg_debug_map ("insert key value pair: %s -> %s; line: %d",
-								key, value, line_number);
+								stripped_key, stripped_value, line_number);
 						g_free (key);
 						g_free (value);
 						key = NULL;
 						value = NULL;
 					} else {
-						func (data->cur_data, key, default_value);
+						func (data->cur_data, stripped_key, default_value);
 						msg_debug_map ("insert key only pair: %s -> %s; line: %d",
-								key, default_value, line_number);
+								stripped_key, default_value, line_number);
 						g_free (key);
 						key = NULL;
 					}
@@ -340,17 +340,17 @@ rspamd_parse_kv_list (
 					if (p - c > 0) {
 						/* Store a single key */
 						MAP_STORE_VALUE;
-						func (data->cur_data, key, value);
+						func (data->cur_data, stripped_key, stripped_value);
 						msg_debug_map ("insert key value pair: %s -> %s",
-								key, value);
+								stripped_key, stripped_value);
 						g_free (key);
 						g_free (value);
 						key = NULL;
 						value = NULL;
 					} else {
-						func (data->cur_data, key, default_value);
+						func (data->cur_data, stripped_key, default_value);
 						msg_debug_map ("insert key only pair: %s -> %s",
-								key, default_value);
+								stripped_key, default_value);
 						g_free (key);
 						key = NULL;
 					}
@@ -408,9 +408,9 @@ rspamd_parse_kv_list (
 			if (p - c > 0) {
 				/* Store a single key */
 				MAP_STORE_KEY;
-				func (data->cur_data, key, default_value);
+				func (data->cur_data, stripped_key, default_value);
 				msg_debug_map ("insert key only pair: %s -> %s",
-						key, default_value);
+						stripped_key, default_value);
 				g_free (key);
 				key = NULL;
 			}
@@ -425,17 +425,17 @@ rspamd_parse_kv_list (
 				if (p - c > 0) {
 					/* Store a single key */
 					MAP_STORE_VALUE;
-					func (data->cur_data, key, value);
+					func (data->cur_data, stripped_key, stripped_value);
 					msg_debug_map ("insert key value pair: %s -> %s",
-							key, value);
+							stripped_key, stripped_value);
 					g_free (key);
 					g_free (value);
 					key = NULL;
 					value = NULL;
 				} else {
-					func (data->cur_data, key, default_value);
+					func (data->cur_data, stripped_key, default_value);
 					msg_debug_map ("insert key only pair: %s -> %s",
-							key, default_value);
+							stripped_key, default_value);
 					g_free (key);
 					key = NULL;
 				}
@@ -741,7 +741,7 @@ rspamd_map_helper_new_hash (struct rspamd_map *map)
 				NULL, 0);
 	}
 
-	htb = rspamd_mempool_alloc0 (pool, sizeof (*htb));
+	htb = rspamd_mempool_alloc0_type(pool, struct rspamd_hash_map_helper);
 	htb->htb = kh_init (rspamd_map_hash);
 	htb->pool = pool;
 	htb->map = map;
@@ -801,7 +801,7 @@ rspamd_map_helper_new_radix (struct rspamd_map *map)
 				NULL, 0);
 	}
 
-	r = rspamd_mempool_alloc0 (pool, sizeof (*r));
+	r = rspamd_mempool_alloc0_type (pool, struct rspamd_radix_map_helper);
 	r->trie = radix_create_compressed_with_pool (pool, name);
 	r->htb = kh_init (rspamd_map_hash);
 	r->pool = pool;
@@ -855,7 +855,7 @@ rspamd_map_helper_new_regexp (struct rspamd_map *map,
 	pool = rspamd_mempool_new (rspamd_mempool_suggest_size (),
 			map->tag, 0);
 
-	re_map = rspamd_mempool_alloc0 (pool, sizeof (*re_map));
+	re_map = rspamd_mempool_alloc0_type (pool, struct rspamd_regexp_map_helper);
 	re_map->pool = pool;
 	re_map->values = g_ptr_array_new ();
 	re_map->regexps = g_ptr_array_new ();
@@ -940,22 +940,34 @@ rspamd_kv_list_fin (struct map_cb_data *data, void **target)
 	struct rspamd_map *map = data->map;
 	struct rspamd_hash_map_helper *htb;
 
-	if (data->cur_data) {
-		htb = (struct rspamd_hash_map_helper *)data->cur_data;
-		msg_info_map ("read hash of %d elements from %s", kh_size (htb->htb),
-				map->name);
-		data->map->traverse_function = rspamd_map_helper_traverse_hash;
-		data->map->nelts = kh_size (htb->htb);
-		data->map->digest = rspamd_cryptobox_fast_hash_final (&htb->hst);
+	if (data->errored) {
+		/* Clean up the current data and do not touch prev data */
+		if (data->cur_data) {
+			msg_info_map ("cleanup unfinished new data as error occurred for %s",
+					map->name);
+			htb = (struct rspamd_hash_map_helper *) data->cur_data;
+			rspamd_map_helper_destroy_hash(htb);
+			data->cur_data = NULL;
+		}
 	}
+	else {
+		if (data->cur_data) {
+			htb = (struct rspamd_hash_map_helper *) data->cur_data;
+			msg_info_map ("read hash of %d elements from %s", kh_size(htb->htb),
+					map->name);
+			data->map->traverse_function = rspamd_map_helper_traverse_hash;
+			data->map->nelts = kh_size (htb->htb);
+			data->map->digest = rspamd_cryptobox_fast_hash_final(&htb->hst);
+		}
 
-	if (target) {
-		*target = data->cur_data;
-	}
+		if (target) {
+			*target = data->cur_data;
+		}
 
-	if (data->prev_data) {
-		htb = (struct rspamd_hash_map_helper *)data->prev_data;
-		rspamd_map_helper_destroy_hash (htb);
+		if (data->prev_data) {
+			htb = (struct rspamd_hash_map_helper *) data->prev_data;
+			rspamd_map_helper_destroy_hash(htb);
+		}
 	}
 }
 
@@ -1000,22 +1012,34 @@ rspamd_radix_fin (struct map_cb_data *data, void **target)
 	struct rspamd_map *map = data->map;
 	struct rspamd_radix_map_helper *r;
 
-	if (data->cur_data) {
-		r = (struct rspamd_radix_map_helper *)data->cur_data;
-		msg_info_map ("read radix trie of %z elements: %s",
-				radix_get_size (r->trie), radix_get_info (r->trie));
-		data->map->traverse_function = rspamd_map_helper_traverse_radix;
-		data->map->nelts = kh_size (r->htb);
-		data->map->digest = rspamd_cryptobox_fast_hash_final (&r->hst);
+	if (data->errored) {
+		/* Clean up the current data and do not touch prev data */
+		if (data->cur_data) {
+			msg_info_map ("cleanup unfinished new data as error occurred for %s",
+					map->name);
+			r = (struct rspamd_radix_map_helper *) data->cur_data;
+			rspamd_map_helper_destroy_radix(r);
+			data->cur_data = NULL;
+		}
 	}
+	else {
+		if (data->cur_data) {
+			r = (struct rspamd_radix_map_helper *) data->cur_data;
+			msg_info_map ("read radix trie of %z elements: %s",
+					radix_get_size(r->trie), radix_get_info(r->trie));
+			data->map->traverse_function = rspamd_map_helper_traverse_radix;
+			data->map->nelts = kh_size (r->htb);
+			data->map->digest = rspamd_cryptobox_fast_hash_final(&r->hst);
+		}
 
-	if (target) {
-		*target = data->cur_data;
-	}
+		if (target) {
+			*target = data->cur_data;
+		}
 
-	if (data->prev_data) {
-		r = (struct rspamd_radix_map_helper *)data->prev_data;
-		rspamd_map_helper_destroy_radix (r);
+		if (data->prev_data) {
+			r = (struct rspamd_radix_map_helper *) data->prev_data;
+			rspamd_map_helper_destroy_radix(r);
+		}
 	}
 }
 
@@ -1095,6 +1119,8 @@ rspamd_re_maps_cache_cleanup_dtor (gpointer ud)
 
 			g_free (basename);
 		}
+
+		g_ptr_array_free (cache_files, TRUE);
 	}
 
 	g_hash_table_unref (cbd->valid_re_hashes);
@@ -1280,9 +1306,9 @@ rspamd_re_map_finalize (struct rspamd_regexp_map_helper *re_map)
 
 	map = re_map->map;
 
-#ifndef __aarch64__
+#if !defined(__aarch64__) && !defined(__powerpc64__)
 	if (!(map->cfg->libs_ctx->crypto_ctx->cpu_config & CPUID_SSSE3)) {
-		msg_info_map ("disable hyperscan for map %s, ssse3 instructons are not supported by CPU",
+		msg_info_map ("disable hyperscan for map %s, ssse3 instructions are not supported by CPU",
 				map->name);
 		return;
 	}
@@ -1489,36 +1515,48 @@ rspamd_glob_list_read_multiple (
 void
 rspamd_regexp_list_fin (struct map_cb_data *data, void **target)
 {
-	struct rspamd_regexp_map_helper *re_map, *old_re_map;
+	struct rspamd_regexp_map_helper *re_map = NULL, *old_re_map;
 	struct rspamd_map *map = data->map;
 
-	if (data->cur_data) {
-		re_map = data->cur_data;
-		rspamd_cryptobox_hash_final (&re_map->hst, re_map->re_digest);
-		memcpy (&data->map->digest, re_map->re_digest, sizeof (data->map->digest));
-		rspamd_re_map_finalize (re_map);
-		msg_info_map ("read regexp list of %ud elements",
-				re_map->regexps->len);
-		data->map->traverse_function = rspamd_map_helper_traverse_regexp;
-		data->map->nelts = kh_size (re_map->htb);
+	if (data->errored) {
+		/* Clean up the current data and do not touch prev data */
+		if (data->cur_data) {
+			msg_info_map ("cleanup unfinished new data as error occurred for %s",
+					map->name);
+			re_map = (struct rspamd_regexp_map_helper *)data->cur_data;
+			rspamd_map_helper_destroy_regexp (re_map);
+			data->cur_data = NULL;
+		}
 	}
+	else {
+		if (data->cur_data) {
+			re_map = data->cur_data;
+			rspamd_cryptobox_hash_final(&re_map->hst, re_map->re_digest);
+			memcpy(&data->map->digest, re_map->re_digest, sizeof(data->map->digest));
+			rspamd_re_map_finalize(re_map);
+			msg_info_map ("read regexp list of %ud elements",
+					re_map->regexps->len);
+			data->map->traverse_function = rspamd_map_helper_traverse_regexp;
+			data->map->nelts = kh_size (re_map->htb);
+		}
 
-	if (target) {
-		*target = data->cur_data;
-	}
+		if (target) {
+			*target = data->cur_data;
+		}
 
-	if (data->prev_data) {
-		old_re_map = data->prev_data;
+		if (data->prev_data) {
+			old_re_map = data->prev_data;
 
 #ifdef WITH_HYPERSCAN
-		if (memcmp (re_map->re_digest, old_re_map->re_digest,
-				sizeof (re_map->re_digest)) != 0) {
-			/* Cleanup old stuff */
-			rspamd_re_map_cache_cleanup_old (old_re_map);
-		}
+			if (re_map && memcmp(re_map->re_digest, old_re_map->re_digest,
+					sizeof(re_map->re_digest)) != 0) {
+				/* Cleanup old stuff */
+				rspamd_re_map_cache_cleanup_old(old_re_map);
+			}
 #endif
 
-		rspamd_map_helper_destroy_regexp (old_re_map);
+			rspamd_map_helper_destroy_regexp(old_re_map);
+		}
 	}
 }
 void
@@ -1866,6 +1904,7 @@ rspamd_cdb_list_read (gchar *chunk,
 		cdb = g_malloc0 (sizeof (struct cdb));
 
 		if (cdb_init (cdb, fd) == -1) {
+			g_free (cdb);
 			msg_err_map ("cannot init cdb map from %s: %s", chunk, strerror (errno));
 
 			return NULL;
@@ -1886,21 +1925,33 @@ rspamd_cdb_list_fin (struct map_cb_data *data, void **target)
 	struct rspamd_map *map = data->map;
 	struct rspamd_cdb_map_helper *cdb_data;
 
-	if (data->cur_data) {
-		cdb_data = (struct rspamd_cdb_map_helper *)data->cur_data;
-		msg_info_map ("read cdb of %Hz size", cdb_data->total_size);
-		data->map->traverse_function = NULL;
-		data->map->nelts = 0;
-		data->map->digest = rspamd_cryptobox_fast_hash_final (&cdb_data->hst);
+	if (data->errored) {
+		/* Clean up the current data and do not touch prev data */
+		if (data->cur_data) {
+			msg_info_map ("cleanup unfinished new data as error occurred for %s",
+					map->name);
+			cdb_data = (struct rspamd_cdb_map_helper *) data->cur_data;
+			rspamd_map_helper_destroy_cdb(cdb_data);
+			data->cur_data = NULL;
+		}
 	}
+	else {
+		if (data->cur_data) {
+			cdb_data = (struct rspamd_cdb_map_helper *) data->cur_data;
+			msg_info_map ("read cdb of %Hz size", cdb_data->total_size);
+			data->map->traverse_function = NULL;
+			data->map->nelts = 0;
+			data->map->digest = rspamd_cryptobox_fast_hash_final(&cdb_data->hst);
+		}
 
-	if (target) {
-		*target = data->cur_data;
-	}
+		if (target) {
+			*target = data->cur_data;
+		}
 
-	if (data->prev_data) {
-		cdb_data = (struct rspamd_cdb_map_helper *)data->prev_data;
-		rspamd_map_helper_destroy_cdb (cdb_data);
+		if (data->prev_data) {
+			cdb_data = (struct rspamd_cdb_map_helper *) data->prev_data;
+			rspamd_map_helper_destroy_cdb(cdb_data);
+		}
 	}
 }
 void
