@@ -406,7 +406,7 @@ rspamd_re_cache_init (struct rspamd_re_cache *cache, struct rspamd_config *cfg)
 		rspamd_regexp_set_cache_id (re, i);
 
 		if (re_class->st == NULL) {
-			(void) !posix_memalign ((void **)&re_class->st, _Alignof (rspamd_cryptobox_hash_state_t),
+			(void) !posix_memalign ((void **)&re_class->st, RSPAMD_ALIGNOF(rspamd_cryptobox_hash_state_t),
 			 		sizeof (*re_class->st));
 			g_assert (re_class->st != NULL);
 			rspamd_cryptobox_hash_init (re_class->st, NULL, 0);
@@ -441,7 +441,7 @@ rspamd_re_cache_init (struct rspamd_re_cache *cache, struct rspamd_config *cfg)
 				sizeof (fl));
 		rspamd_cryptobox_hash_update (&st_global, (const guchar *) &fl,
 				sizeof (fl));
-		/* Numberic order */
+		/* Numeric order */
 		rspamd_cryptobox_hash_update (re_class->st, (const guchar *)&i,
 				sizeof (i));
 		rspamd_cryptobox_hash_update (&st_global, (const guchar *)&i,
@@ -570,6 +570,7 @@ rspamd_re_cache_check_lua_condition (struct rspamd_task *task,
 		msg_warn_task ("cannot call for re_cache_check_lua_condition for re %s: %e",
 				rspamd_regexp_get_pattern (re), err);
 		g_error_free (err);
+		lua_settop (L, text_pos - 1);
 
 		return TRUE;
 	}
@@ -1154,8 +1155,8 @@ rspamd_re_cache_exec_re (struct rspamd_task *task,
 	case RSPAMD_RE_HEADER:
 	case RSPAMD_RE_RAWHEADER:
 		/* Get list of specified headers */
-		rh = rspamd_message_get_header_array (task,
-				re_class->type_data);
+		rh = rspamd_message_get_header_array(task,
+				re_class->type_data, FALSE);
 
 		if (rh) {
 			ret = rspamd_re_cache_process_headers_list (task, rt, re,
@@ -1177,8 +1178,8 @@ rspamd_re_cache_exec_re (struct rspamd_task *task,
 		break;
 	case RSPAMD_RE_MIMEHEADER:
 		PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, parts), i, mime_part) {
-			rh = rspamd_message_get_header_from_hash (mime_part->raw_headers,
-					re_class->type_data);
+			rh = rspamd_message_get_header_from_hash(mime_part->raw_headers,
+					re_class->type_data, FALSE);
 
 			if (rh) {
 				ret += rspamd_re_cache_process_headers_list (task, rt, re,
@@ -1224,8 +1225,8 @@ rspamd_re_cache_exec_re (struct rspamd_task *task,
 							raw = TRUE;
 						}
 
-						in = text_part->utf_content->data;
-						len = text_part->utf_content->len;
+						in = text_part->utf_content.begin;
+						len = text_part->utf_content.len;
 					}
 				}
 
@@ -1345,7 +1346,7 @@ rspamd_re_cache_exec_re (struct rspamd_task *task,
 		 * of the body content.
 		 */
 
-		rh = rspamd_message_get_header_array (task, "Subject");
+		rh = rspamd_message_get_header_array(task, "Subject", FALSE);
 
 		if (rh) {
 			scvec[0] = (guchar *)rh->decoded;
@@ -1743,7 +1744,7 @@ static gchar *
 rspamd_re_cache_hs_pattern_from_pcre (rspamd_regexp_t *re)
 {
 	/*
-	 * Workaroung for bug in ragel 7.0.0.11
+	 * Workaround for bug in ragel 7.0.0.11
 	 * https://github.com/intel/hyperscan/issues/133
 	 */
 	const gchar *pat = rspamd_regexp_get_pattern (re);
@@ -1862,12 +1863,19 @@ struct rspamd_re_cache_hs_compile_cbdata {
 
 static void
 rspamd_re_cache_compile_err (EV_P_ ev_timer *w, GError *err,
-		struct rspamd_re_cache_hs_compile_cbdata *cbdata)
+		struct rspamd_re_cache_hs_compile_cbdata *cbdata, bool is_fatal)
 {
-	ev_timer_stop (EV_A_ w);
 	cbdata->cb (cbdata->total, err, cbdata->cbd);
-	g_free (w);
-	g_free (cbdata);
+
+	if (is_fatal) {
+		ev_timer_stop(EV_A_ w);
+		g_free(w);
+		g_free(cbdata);
+	}
+	else {
+		/* Continue compilation */
+		ev_timer_again(EV_A_ w);
+	}
 	g_error_free (err);
 }
 
@@ -1885,15 +1893,16 @@ rspamd_re_cache_compile_timer_cb (EV_P_ ev_timer *w, int revents )
 	rspamd_cryptobox_fast_hash_state_t crc_st;
 	guint64 crc;
 	rspamd_regexp_t *re;
-	hs_compile_error_t *hs_errors;
+	hs_compile_error_t *hs_errors = NULL;
 	guint *hs_flags = NULL;
 	const hs_expr_ext_t **hs_exts = NULL;
 	gchar **hs_pats = NULL;
-	gchar *hs_serialized;
+	gchar *hs_serialized = NULL;
 	gsize serialized_len;
 	struct iovec iov[7];
 	struct rspamd_re_cache *cache;
 	GError *err;
+	pid_t our_pid = getpid ();
 
 	cache = cbdata->cache;
 
@@ -1917,7 +1926,7 @@ rspamd_re_cache_compile_timer_cb (EV_P_ ev_timer *w, int revents )
 
 		/* Read number of regexps */
 		g_assert (fd != -1);
-		lseek (fd, RSPAMD_HS_MAGIC_LEN + sizeof (cache->plt), SEEK_SET);
+		g_assert (lseek (fd, RSPAMD_HS_MAGIC_LEN + sizeof (cache->plt), SEEK_SET) != -1);
 		g_assert (read (fd, &n, sizeof (n)) == sizeof (n));
 		close (fd);
 
@@ -1946,14 +1955,14 @@ rspamd_re_cache_compile_timer_cb (EV_P_ ev_timer *w, int revents )
 		return;
 	}
 
-	rspamd_snprintf (path, sizeof (path), "%s%c%s.hs.new", cbdata->cache_dir,
-			G_DIR_SEPARATOR, re_class->hash);
+	rspamd_snprintf (path, sizeof (path), "%s%c%s.%P.hs.new", cbdata->cache_dir,
+			G_DIR_SEPARATOR, re_class->hash, our_pid);
 	fd = open (path, O_CREAT|O_TRUNC|O_EXCL|O_WRONLY, 00600);
 
 	if (fd == -1) {
 		err = g_error_new (rspamd_re_cache_quark (), errno,
 				"cannot open file %s: %s", path, strerror (errno));
-		rspamd_re_cache_compile_err (EV_A_ w, err, cbdata);
+		rspamd_re_cache_compile_err (EV_A_ w, err, cbdata, false);
 		return;
 	}
 
@@ -2044,8 +2053,24 @@ rspamd_re_cache_compile_timer_cb (EV_P_ ev_timer *w, int revents )
 	/* Adjust real re number */
 	n = i;
 
+#define CLEANUP_ALLOCATED(is_err) do {    \
+    g_free (hs_flags);                    \
+    g_free (hs_ids);                    \
+    for (guint j = 0; j < i; j ++) {    \
+        g_free (hs_pats[j]);            \
+    }                                    \
+    g_free (hs_pats);                    \
+    g_free (hs_exts);                    \
+    if (is_err) {                         \
+        close (fd);                            \
+        unlink (path);                        \
+        if (hs_errors) hs_free_compile_error (hs_errors); \
+    }                                        \
+} while(0)
+
 	if (n > 0) {
 		/* Create the hs tree */
+		hs_errors = NULL;
 		if (hs_compile_ext_multi ((const char **)hs_pats,
 				hs_flags,
 				hs_ids,
@@ -2056,35 +2081,15 @@ rspamd_re_cache_compile_timer_cb (EV_P_ ev_timer *w, int revents )
 				&test_db,
 				&hs_errors) != HS_SUCCESS) {
 
-
-			g_free (hs_flags);
-			g_free (hs_ids);
-
-			for (guint j = 0; j < i; j ++) {
-				g_free (hs_pats[j]);
-			}
-
-			g_free (hs_pats);
-			g_free (hs_exts);
-			close (fd);
-			unlink (path);
-			hs_free_compile_error (hs_errors);
-
 			err = g_error_new (rspamd_re_cache_quark (), EINVAL,
 					"cannot create tree of regexp when processing '%s': %s",
 					hs_pats[hs_errors->expression], hs_errors->message);
-			rspamd_re_cache_compile_err (EV_A_ w, err, cbdata);
+			CLEANUP_ALLOCATED(true);
+			rspamd_re_cache_compile_err (EV_A_ w, err, cbdata, false);
 
 			return;
 		}
 
-		for (guint j = 0; j < i; j ++) {
-			g_free (hs_pats[j]);
-		}
-
-		g_free (hs_pats);
-		g_free (hs_exts);
-
 		if (hs_serialize_database (test_db, &hs_serialized,
 				&serialized_len) != HS_SUCCESS) {
 			err = g_error_new (rspamd_re_cache_quark (),
@@ -2092,13 +2097,9 @@ rspamd_re_cache_compile_timer_cb (EV_P_ ev_timer *w, int revents )
 					"cannot serialize tree of regexp for %s",
 					re_class->hash);
 
-			close (fd);
-			unlink (path);
-			g_free (hs_ids);
-			g_free (hs_flags);
+			CLEANUP_ALLOCATED(true);
 			hs_free_database (test_db);
-
-			rspamd_re_cache_compile_err (EV_A_ w, err, cbdata);
+			rspamd_re_cache_compile_err (EV_A_ w, err, cbdata, false);
 			return;
 		}
 
@@ -2149,13 +2150,11 @@ rspamd_re_cache_compile_timer_cb (EV_P_ ev_timer *w, int revents )
 					errno,
 					"cannot serialize tree of regexp to %s: %s",
 					path, strerror (errno));
-			close (fd);
-			unlink (path);
-			g_free (hs_ids);
-			g_free (hs_flags);
+
+			CLEANUP_ALLOCATED(true);
 			g_free (hs_serialized);
 
-			rspamd_re_cache_compile_err (EV_A_ w, err, cbdata);
+			rspamd_re_cache_compile_err (EV_A_ w, err, cbdata, false);
 			return;
 		}
 
@@ -2179,13 +2178,10 @@ rspamd_re_cache_compile_timer_cb (EV_P_ ev_timer *w, int revents )
 		}
 
 		cbdata->total += n;
-
-		g_free (hs_serialized);
-		g_free (hs_ids);
-		g_free (hs_flags);
+		CLEANUP_ALLOCATED(false);
 
 		/* Now rename temporary file to the new .hs file */
-		rspamd_snprintf (npath, sizeof (path), "%s%c%s.hs", cbdata->cache_dir,
+		rspamd_snprintf (npath, sizeof (npath), "%s%c%s.hs", cbdata->cache_dir,
 				G_DIR_SEPARATOR, re_class->hash);
 
 		if (rename (path, npath) == -1) {
@@ -2196,7 +2192,7 @@ rspamd_re_cache_compile_timer_cb (EV_P_ ev_timer *w, int revents )
 			unlink (path);
 			close (fd);
 
-			rspamd_re_cache_compile_err (EV_A_ w, err, cbdata);
+			rspamd_re_cache_compile_err (EV_A_ w, err, cbdata, false);
 			return;
 		}
 
@@ -2211,9 +2207,8 @@ rspamd_re_cache_compile_timer_cb (EV_P_ ev_timer *w, int revents )
 				(gint)g_hash_table_size (re_class->re),
 				path);
 
-		unlink (path);
-		close (fd);
-		rspamd_re_cache_compile_err (EV_A_ w, err, cbdata);
+		CLEANUP_ALLOCATED(true);
+		rspamd_re_cache_compile_err (EV_A_ w, err, cbdata, false);
 
 		return;
 	}
@@ -2290,10 +2285,18 @@ rspamd_re_cache_is_valid_hyperscan_file (struct rspamd_re_cache *cache,
 	len = strlen (path);
 
 	if (len < sizeof (rspamd_cryptobox_HASHBYTES + 3)) {
+		if (!silent)  {
+			msg_err_re_cache ("cannot open hyperscan cache file %s: too short filename",
+					path);
+		}
 		return FALSE;
 	}
 
 	if (memcmp (path + len - 3, ".hs", 3) != 0) {
+		if (!silent)  {
+			msg_err_re_cache ("cannot open hyperscan cache file %s: not ending with .hs",
+					path);
+		}
 		return FALSE;
 	}
 
@@ -2388,7 +2391,7 @@ rspamd_re_cache_is_valid_hyperscan_file (struct rspamd_re_cache *cache,
 
 				p = map + RSPAMD_HS_MAGIC_LEN + sizeof (test_plt);
 				end = map + len;
-				n = *(gint *)p;
+				memcpy(&n, p, sizeof(n));
 				p += sizeof (gint);
 
 				if (n <= 0 || 2 * n * sizeof (gint) + /* IDs + flags */
