@@ -301,6 +301,11 @@ gboolean rspamd_symcache_set_symbol_flags (struct rspamd_symcache *cache,
 guint rspamd_symcache_get_symbol_flags (struct rspamd_symcache *cache,
 										const gchar *symbol);
 
+void rspamd_symcache_get_symbol_details(struct rspamd_symcache *cache,
+								 const gchar *symbol,
+								 ucl_object_t *this_sym_ucl);
+
+
 /**
  * Process settings for task
  * @param task
@@ -523,7 +528,7 @@ gboolean rspamd_symcache_is_item_allowed (struct rspamd_task *task,
 										  gboolean exec_only);
 
 /**
- * Returns symbcache item flags
+ * Returns symcache item flags
  * @param item
  * @return
  */
@@ -542,7 +547,7 @@ const gchar* rspamd_symcache_item_name (struct rspamd_symcache_item *item);
 const struct rspamd_symcache_item_stat *
 		rspamd_symcache_item_stat (struct rspamd_symcache_item *item);
 /**
- * Returns if an item is enabled (for virutal it also means that parent should be enabled)
+ * Returns if an item is enabled (for virtual it also means that parent should be enabled)
  * @param item
  * @return
  */
