@@ -31,7 +31,9 @@
 #define SPF_ALL "all"
 #define SPF_A "a"
 #define SPF_IP4 "ip4"
+#define SPF_IP4_ALT "ipv4"
 #define SPF_IP6 "ip6"
+#define SPF_IP6_ALT "ipv6"
 #define SPF_PTR "ptr"
 #define SPF_MX "mx"
 #define SPF_EXISTS "exists"
@@ -42,7 +44,7 @@
 struct spf_resolved_element {
 	GPtrArray *elts;
 	gchar *cur_domain;
-	gboolean redirected; /* Ingnore level, it's redirected */
+	gboolean redirected; /* Ignore level, it's redirected */
 };
 
 struct spf_record {
@@ -55,6 +57,7 @@ struct spf_record {
 	/* Array of struct spf_resolved_element */
 	const gchar *sender;
 	const gchar *sender_domain;
+	const gchar *top_record;
 	gchar *local_part;
 	struct rspamd_task *task;
 	spf_cb_t callback;
@@ -378,6 +381,7 @@ rspamd_flatten_record_dtor (struct spf_resolved *r)
 		g_free (addr->spf_string);
 	}
 
+	g_free (r->top_record);
 	g_free (r->domain);
 	g_array_free (r->elts, TRUE);
 	g_free (r);
@@ -498,6 +502,7 @@ rspamd_spf_record_flatten (struct spf_record *rec)
 	/* Not precise but okay */
 	res->timestamp = rec->task->task_timestamp;
 	res->digest = mum_hash_init (0xa4aa40bbeec59e2bULL);
+	res->top_record = g_strdup (rec->top_record);
 	REF_INIT_RETAIN (res, rspamd_flatten_record_dtor);
 
 	if (rec->resolved) {
@@ -565,8 +570,11 @@ rspamd_spf_record_postprocess (struct spf_resolved *rec, struct rspamd_task *tas
 		if (cur_addr->flags & RSPAMD_SPF_FLAG_IPV6) {
 			guint64 t[3];
 
+			/*
+			 * Fill hash entry for ipv6 addr with 2 int64 from ipv6 address,
+			 * the remaining int64 has mech + mask
+			 */
 			memcpy (t, cur_addr->addr6, sizeof (guint64) * 2);
-			t[2] = 0;
 			t[2] = ((guint64) (cur_addr->mech)) << 48u;
 			t[2] |= cur_addr->m.dual.mask_v6;
 
@@ -599,6 +607,7 @@ rspamd_spf_maybe_return (struct spf_record *rec)
 {
 	struct spf_resolved *flat;
 	struct rspamd_task *task = rec->task;
+	bool cached = false;
 
 	if (rec->requests_inflight == 0 && !rec->done) {
 		flat = rspamd_spf_record_flatten (rec);
@@ -612,16 +621,26 @@ rspamd_spf_maybe_return (struct spf_record *rec)
 						spf_record_ref (flat),
 						flat->timestamp, flat->ttl);
 
-				msg_info_task ("stored record for %s (0x%xuL) in LRU cache for %d seconds, "
+				msg_info_task ("stored SPF record for %s (0x%xuL) in LRU cache for %d seconds, "
 							   "%d/%d elements in the cache",
 						flat->domain,
 						flat->digest,
 						flat->ttl,
 						rspamd_lru_hash_size (spf_lib_ctx->spf_hash),
 						rspamd_lru_hash_capacity (spf_lib_ctx->spf_hash));
+				cached = true;
 			}
 		}
 
+		if (!cached) {
+			/* Still write a log line */
+			msg_info_task ("not stored SPF record for %s (0x%xuL) in LRU cache; flags=%d; ttl=%d",
+					flat->domain,
+					flat->digest,
+					flat->flags,
+					flat->ttl);
+		}
+
 		rec->callback (flat, rec->task, rec->cbdata);
 		spf_record_unref (flat);
 		rec->done = TRUE;
@@ -771,36 +790,44 @@ spf_record_addr_set (struct spf_addr *addr, gboolean allow_any)
 
 static gboolean
 spf_process_txt_record (struct spf_record *rec, struct spf_resolved_element *resolved,
-		struct rdns_reply *reply)
+		struct rdns_reply *reply, struct rdns_reply_entry **pselected)
 {
 	struct rdns_reply_entry *elt, *selected = NULL;
 	gboolean ret = FALSE;
 
 	/*
-	 * We prefer spf version 1 as other records are mostly likely garbadge
+	 * We prefer spf version 1 as other records are mostly likely garbage
 	 * or incorrect records (e.g. spf2 records)
 	 */
 	LL_FOREACH (reply->entries, elt) {
-		if (strncmp (elt->content.txt.data, "v=spf1", sizeof ("v=spf1") - 1)
+		if (elt->type == RDNS_REQUEST_TXT) {
+			if (strncmp(elt->content.txt.data, "v=spf1", sizeof("v=spf1") - 1)
 				== 0) {
-			selected = elt;
-			rspamd_mempool_set_variable (rec->task->task_pool,
-					RSPAMD_MEMPOOL_SPF_RECORD,
-					rspamd_mempool_strdup (rec->task->task_pool,
-							elt->content.txt.data), NULL);
-			break;
+				selected = elt;
+
+				if (pselected != NULL) {
+					*pselected = selected;
+				}
+
+				break;
+			}
 		}
 	}
 
 	if (!selected) {
 		LL_FOREACH (reply->entries, elt) {
-			if (start_spf_parse (rec, resolved, elt->content.txt.data)) {
-				ret = TRUE;
-				rspamd_mempool_set_variable (rec->task->task_pool,
-						RSPAMD_MEMPOOL_SPF_RECORD,
-						rspamd_mempool_strdup (rec->task->task_pool,
-								elt->content.txt.data), NULL);
-				break;
+			/*
+			 * Rubbish spf record? Let's still try to process it, but merely for
+			 * TXT RRs
+			 */
+			if (elt->type == RDNS_REQUEST_TXT) {
+				if (start_spf_parse(rec, resolved, elt->content.txt.data)) {
+					ret = TRUE;
+					if (pselected != NULL) {
+						*pselected = elt;
+					}
+					break;
+				}
 			}
 		}
 	}
@@ -820,15 +847,32 @@ spf_record_dns_callback (struct rdns_reply *reply, gpointer arg)
 	struct spf_addr *addr;
 	struct spf_record *rec;
 	const struct rdns_request_name *req_name;
+	bool truncated = false;
 
 	rec = cb->rec;
 	task = rec->task;
 
 	cb->rec->requests_inflight--;
 	addr = cb->addr;
+	req_name = rdns_request_get_name (reply->request, NULL);
 
-	if (reply->code == RDNS_RC_NOERROR) {
-		req_name = rdns_request_get_name (reply->request, NULL);
+	if (reply->flags & RDNS_TRUNCATED) {
+		/* Do not process truncated DNS replies */
+		truncated = true;
+
+		if (req_name) {
+			msg_warn_spf ("got a truncated record when trying to resolve %s (%s type) for SPF domain %s",
+					req_name->name, rdns_str_from_type(req_name->type),
+					rec->sender_domain);
+		}
+		else {
+			msg_warn_spf ("got a truncated record when trying to resolve ??? "
+						  "(internal error) for SPF domain %s",
+					rec->sender_domain);
+		}
+	}
+
+	if (reply->code == RDNS_RC_NOERROR && !truncated) {
 
 		LL_FOREACH (reply->entries, elt_data) {
 			/* Adjust ttl if a resolved record has lower ttl than spf record itself */
@@ -922,7 +966,7 @@ spf_record_dns_callback (struct rdns_reply *reply, gpointer arg)
 									reply->entries[0].content.txt.data);
 						}
 
-						if (!spf_process_txt_record (rec, cb->resolved, reply)) {
+						if (!spf_process_txt_record (rec, cb->resolved, reply, NULL)) {
 							cb->addr->flags |= RSPAMD_SPF_FLAG_PERMFAIL;
 						}
 					}
@@ -938,7 +982,7 @@ spf_record_dns_callback (struct rdns_reply *reply, gpointer arg)
 						}
 
 						cb->addr->flags |= RSPAMD_SPF_FLAG_RESOLVED;
-						spf_process_txt_record (rec, cb->resolved, reply);
+						spf_process_txt_record (rec, cb->resolved, reply, NULL);
 					}
 					goto end;
 
@@ -964,7 +1008,7 @@ spf_record_dns_callback (struct rdns_reply *reply, gpointer arg)
 			case SPF_RESOLVE_MX:
 				if (!(cb->addr->flags & RSPAMD_SPF_FLAG_RESOLVED)) {
 					cb->addr->flags |= RSPAMD_SPF_FLAG_PERMFAIL;
-					msg_debug_spf (
+					msg_info_spf (
 							"spf error for domain %s: cannot find MX"
 							" record for %s: %s",
 							cb->rec->sender_domain,
@@ -976,7 +1020,7 @@ spf_record_dns_callback (struct rdns_reply *reply, gpointer arg)
 			case SPF_RESOLVE_A:
 				if (!(cb->addr->flags & RSPAMD_SPF_FLAG_RESOLVED)) {
 					cb->addr->flags |= RSPAMD_SPF_FLAG_PERMFAIL;
-					msg_debug_spf (
+					msg_info_spf (
 							"spf error for domain %s: cannot resolve A"
 							" record for %s: %s",
 							cb->rec->sender_domain,
@@ -991,7 +1035,7 @@ spf_record_dns_callback (struct rdns_reply *reply, gpointer arg)
 			case SPF_RESOLVE_AAA:
 				if (!(cb->addr->flags & RSPAMD_SPF_FLAG_RESOLVED)) {
 					cb->addr->flags |= RSPAMD_SPF_FLAG_PERMFAIL;
-					msg_debug_spf (
+					msg_info_spf (
 							"spf error for domain %s: cannot resolve AAAA"
 							" record for %s: %s",
 							cb->rec->sender_domain,
@@ -1004,7 +1048,7 @@ spf_record_dns_callback (struct rdns_reply *reply, gpointer arg)
 				break;
 			case SPF_RESOLVE_PTR:
 				if (!(cb->addr->flags & RSPAMD_SPF_FLAG_RESOLVED)) {
-					msg_debug_spf (
+					msg_info_spf (
 							"spf error for domain %s: cannot resolve PTR"
 							" record for %s: %s",
 							cb->rec->sender_domain,
@@ -1018,7 +1062,7 @@ spf_record_dns_callback (struct rdns_reply *reply, gpointer arg)
 			case SPF_RESOLVE_REDIRECT:
 				if (!(cb->addr->flags & RSPAMD_SPF_FLAG_RESOLVED)) {
 					cb->addr->flags |= RSPAMD_SPF_FLAG_PERMFAIL;
-					msg_debug_spf (
+					msg_info_spf (
 							"spf error for domain %s: cannot resolve REDIRECT"
 							" record for %s: %s",
 							cb->rec->sender_domain,
@@ -1029,7 +1073,7 @@ spf_record_dns_callback (struct rdns_reply *reply, gpointer arg)
 				break;
 			case SPF_RESOLVE_INCLUDE:
 				if (!(cb->addr->flags & RSPAMD_SPF_FLAG_RESOLVED)) {
-					msg_debug_spf (
+					msg_info_spf (
 							"spf error for domain %s: cannot resolve INCLUDE"
 							" record for %s: %s",
 							cb->rec->sender_domain,
@@ -1044,8 +1088,8 @@ spf_record_dns_callback (struct rdns_reply *reply, gpointer arg)
 			case SPF_RESOLVE_EXISTS:
 				if (!(cb->addr->flags & RSPAMD_SPF_FLAG_RESOLVED)) {
 					msg_debug_spf (
-							"spf error for domain %s: cannot resolve EXISTS"
-							" record for %s: %s",
+							"spf macro resolution for domain %s: cannot resolve EXISTS"
+							" macro for %s: %s",
 							cb->rec->sender_domain,
 							cb->resolved->cur_domain,
 							rdns_strerror (reply->code));
@@ -1090,7 +1134,7 @@ parse_spf_domain_mask (struct spf_record *rec, struct spf_addr *addr,
 		parse_ipv4_mask,
 		parse_second_slash,
 		parse_ipv6_mask,
-		skip_garbadge
+		skip_garbage
 	} state = 0;
 	const gchar *p = addr->spf_string, *host, *c;
 	gchar *hostbuf;
@@ -1139,7 +1183,7 @@ parse_spf_domain_mask (struct spf_record *rec, struct spf_addr *addr,
 					state = parse_ipv4_mask;
 				}
 				else {
-					state = skip_garbadge;
+					state = skip_garbage;
 				}
 				cur_mask = 0;
 				break;
@@ -1172,7 +1216,7 @@ parse_spf_domain_mask (struct spf_record *rec, struct spf_addr *addr,
 				}
 				p++;
 				break;
-			case skip_garbadge:
+			case skip_garbage:
 				p++;
 				break;
 		}
@@ -1777,7 +1821,7 @@ expand_spf_macro (struct spf_record *rec, struct spf_resolved_element *resolved,
 	gchar *c, *new, *tmp, delim = '.';
 	gsize len = 0, slen = 0, macro_len = 0;
 	gint state = 0, ndelim = 0;
-	gchar ip_buf[INET6_ADDRSTRLEN + 1];
+	gchar ip_buf[64 + 1]; /* cannot use INET6_ADDRSTRLEN as we use ptr lookup */
 	gboolean need_expand = FALSE, reversed;
 	struct rspamd_task *task;
 
@@ -1830,7 +1874,7 @@ expand_spf_macro (struct spf_record *rec, struct spf_resolved_element *resolved,
 			/* Read macro name */
 			switch (g_ascii_tolower (*p)) {
 			case 'i':
-				len += INET6_ADDRSTRLEN - 1;
+				len += sizeof (ip_buf) - 1;
 				break;
 			case 's':
 				if (rec->sender) {
@@ -1970,10 +2014,40 @@ expand_spf_macro (struct spf_record *rec, struct spf_resolved_element *resolved,
 			switch (g_ascii_tolower (*p)) {
 			case 'i':
 				if (task->from_addr) {
-					macro_len = rspamd_strlcpy (ip_buf,
-							rspamd_inet_address_to_string (task->from_addr),
-							sizeof (ip_buf));
-					macro_value = ip_buf;
+					if (rspamd_inet_address_get_af (task->from_addr) == AF_INET) {
+						macro_len = rspamd_strlcpy (ip_buf,
+								rspamd_inet_address_to_string (task->from_addr),
+								sizeof (ip_buf));
+						macro_value = ip_buf;
+					}
+					else if (rspamd_inet_address_get_af (task->from_addr) == AF_INET6) {
+						/* See #3625 for details */
+						socklen_t slen;
+						struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)
+								rspamd_inet_address_get_sa (task->from_addr, &slen);
+
+						/* Expand IPv6 address */
+#define IPV6_OCTET(x) bytes[(x)] >> 4, bytes[(x)] & 0xF
+						unsigned char *bytes = (unsigned char *)&sin6->sin6_addr;
+						macro_len = rspamd_snprintf (ip_buf, sizeof (ip_buf),
+								"%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd."
+								"%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd.%xd",
+								IPV6_OCTET(0), IPV6_OCTET(1),
+								IPV6_OCTET(2), IPV6_OCTET(3),
+								IPV6_OCTET(4), IPV6_OCTET(5),
+								IPV6_OCTET(6), IPV6_OCTET(7),
+								IPV6_OCTET(8), IPV6_OCTET(9),
+								IPV6_OCTET(10), IPV6_OCTET(11),
+								IPV6_OCTET(12), IPV6_OCTET(13),
+								IPV6_OCTET(14), IPV6_OCTET(15));
+						macro_value = ip_buf;
+#undef IPV6_OCTET
+					}
+					else {
+						macro_len = rspamd_snprintf (ip_buf, sizeof (ip_buf),
+								"127.0.0.1");
+						macro_value = ip_buf;
+					}
 				}
 				else {
 					macro_len = rspamd_snprintf (ip_buf, sizeof (ip_buf),
@@ -2161,16 +2235,19 @@ spf_process_element (struct spf_record *rec,
 			break;
 		case 'i':
 			/* include or ip4 */
-			if (g_ascii_strncasecmp (begin, SPF_IP4,
-					sizeof (SPF_IP4) - 1) == 0) {
+			if (g_ascii_strncasecmp (begin, SPF_IP4, sizeof (SPF_IP4) - 1) == 0) {
 				res = parse_spf_ip4 (rec, addr);
 			}
-			else if (g_ascii_strncasecmp (begin, SPF_INCLUDE,
-					sizeof (SPF_INCLUDE) - 1) == 0) {
+			else if (g_ascii_strncasecmp (begin, SPF_INCLUDE, sizeof (SPF_INCLUDE) - 1) == 0) {
 				res = parse_spf_include (rec, addr);
 			}
-			else if (g_ascii_strncasecmp (begin, SPF_IP6, sizeof (SPF_IP6) -
-														  1) == 0) {
+			else if (g_ascii_strncasecmp (begin, SPF_IP6, sizeof (SPF_IP6) - 1) == 0) {
+				res = parse_spf_ip6 (rec, addr);
+			}
+			else if (g_ascii_strncasecmp (begin, SPF_IP4_ALT, sizeof (SPF_IP4_ALT) - 1) == 0) {
+				res = parse_spf_ip4 (rec, addr);
+			}
+			else if (g_ascii_strncasecmp (begin, SPF_IP6_ALT, sizeof (SPF_IP6_ALT) - 1) == 0) {
 				res = parse_spf_ip6 (rec, addr);
 			}
 			else {
@@ -2374,30 +2451,46 @@ spf_dns_callback (struct rdns_reply *reply, gpointer arg)
 
 	rec->requests_inflight--;
 
-	if (reply->code == RDNS_RC_NOERROR) {
-		resolved = rspamd_spf_new_addr_list (rec, rec->sender_domain);
-		if (rec->resolved->len == 1) {
-			/* Top level resolved element */
-			rec->ttl = reply->entries->ttl;
-		}
-	}
-	else if ((reply->code == RDNS_RC_NOREC || reply->code == RDNS_RC_NXDOMAIN)
-			&& rec->dns_requests == 0) {
-		resolved = rspamd_spf_new_addr_list (rec, rec->sender_domain);
-		addr = g_malloc0 (sizeof(*addr));
-		addr->flags |= RSPAMD_SPF_FLAG_NA;
-		g_ptr_array_insert (resolved->elts, 0, addr);
-	}
-	else if (reply->code != RDNS_RC_NOREC && reply->code != RDNS_RC_NXDOMAIN
-			&& rec->dns_requests == 0) {
-		resolved = rspamd_spf_new_addr_list (rec, rec->sender_domain);
-		addr = g_malloc0 (sizeof(*addr));
+	if (reply->flags & RDNS_TRUNCATED) {
+		msg_warn_spf ("got a truncated record when trying to resolve TXT record for %s",
+				rec->sender_domain);
+		resolved = rspamd_spf_new_addr_list(rec, rec->sender_domain);
+		addr = g_malloc0(sizeof(*addr));
 		addr->flags |= RSPAMD_SPF_FLAG_TEMPFAIL;
-		g_ptr_array_insert (resolved->elts, 0, addr);
+		g_ptr_array_insert(resolved->elts, 0, addr);
+
+		rspamd_spf_maybe_return (rec);
+
+		return;
+	}
+	else {
+		if (reply->code == RDNS_RC_NOERROR) {
+			resolved = rspamd_spf_new_addr_list(rec, rec->sender_domain);
+			if (rec->resolved->len == 1) {
+				/* Top level resolved element */
+				rec->ttl = reply->entries->ttl;
+			}
+		}
+		else if ((reply->code == RDNS_RC_NOREC || reply->code == RDNS_RC_NXDOMAIN)
+				 && rec->dns_requests == 0) {
+			resolved = rspamd_spf_new_addr_list(rec, rec->sender_domain);
+			addr = g_malloc0(sizeof(*addr));
+			addr->flags |= RSPAMD_SPF_FLAG_NA;
+			g_ptr_array_insert(resolved->elts, 0, addr);
+		}
+		else if (reply->code != RDNS_RC_NOREC && reply->code != RDNS_RC_NXDOMAIN
+				 && rec->dns_requests == 0) {
+			resolved = rspamd_spf_new_addr_list(rec, rec->sender_domain);
+			addr = g_malloc0(sizeof(*addr));
+			addr->flags |= RSPAMD_SPF_FLAG_TEMPFAIL;
+			g_ptr_array_insert(resolved->elts, 0, addr);
+		}
 	}
 
 	if (resolved) {
-		if (!spf_process_txt_record (rec, resolved, reply)) {
+		struct rdns_reply_entry *selected = NULL;
+
+		if (!spf_process_txt_record (rec, resolved, reply, &selected)) {
 			resolved = g_ptr_array_index(rec->resolved, 0);
 
 			if (rec->resolved->len > 1) {
@@ -2422,6 +2515,13 @@ spf_dns_callback (struct rdns_reply *reply, gpointer arg)
 				g_ptr_array_insert (resolved->elts, 0, addr);
 			}
 		}
+		else {
+			rec->top_record = rspamd_mempool_strdup(rec->task->task_pool,
+					selected->content.txt.data);
+			rspamd_mempool_set_variable(rec->task->task_pool,
+					RSPAMD_MEMPOOL_SPF_RECORD,
+					(gpointer)rec->top_record, NULL);
+		}
 	}
 
 	rspamd_spf_maybe_return (rec);
@@ -2521,6 +2621,13 @@ rspamd_spf_resolve (struct rspamd_task *task, spf_cb_t callback,
 
 		if (cached) {
 			cached->flags |= RSPAMD_SPF_FLAG_CACHED;
+
+			if (cached->top_record) {
+				rspamd_mempool_set_variable(task->task_pool,
+						RSPAMD_MEMPOOL_SPF_RECORD,
+						rspamd_mempool_strdup (task->task_pool,
+								cached->top_record), NULL);
+			}
 			callback (cached, task, cbdata);
 
 			return TRUE;
