@@ -21,11 +21,10 @@
 #include "message.h"
 #include "lua/lua_common.h"
 #include "email_addr.h"
-#include "composites.h"
+#include "src/libserver/composites/composites.h"
 #include "stat_api.h"
 #include "unix-std.h"
 #include "utlist.h"
-#include "contrib/zstd/zstd.h"
 #include "libserver/mempool_vars_internal.h"
 #include "libserver/cfg_file_private.h"
 #include "libmime/lang_detection.h"
@@ -41,6 +40,12 @@
 
 #include <math.h>
 
+#ifdef SYS_ZSTD
+#  include "zstd.h"
+#else
+#  include "contrib/zstd/zstd.h"
+#endif
+
 __KHASH_IMPL (rspamd_req_headers_hash, static inline,
 		rspamd_ftok_t *, struct rspamd_request_header_chain *, 1,
 				rspamd_ftok_icase_hash, rspamd_ftok_icase_equal)
@@ -125,7 +130,7 @@ rspamd_task_new (struct rspamd_worker *worker,
 static void
 rspamd_task_reply (struct rspamd_task *task)
 {
-	const ev_tstamp write_timeout = 2.0;
+	const ev_tstamp write_timeout = 5.0;
 
 	if (task->fin_callback) {
 		task->fin_callback (task, task->fin_arg);
@@ -986,7 +991,7 @@ rspamd_task_log_check_condition (struct rspamd_task *task,
 		}
 		break;
 	case RSPAMD_LOG_USER:
-		if (task->user) {
+		if (task->auth_user) {
 			ret = TRUE;
 		}
 		break;
@@ -1120,7 +1125,7 @@ rspamd_task_log_metric_res (struct rspamd_task *task,
 			symbuf = rspamd_fstring_sized_new (128);
 			sorted_symbols = g_ptr_array_sized_new (kh_size (mres->symbols));
 
-			kh_foreach_value_ptr (mres->symbols, sym, {
+			kh_foreach_value (mres->symbols, sym, {
 				if (!(sym->flags & RSPAMD_SYMBOL_RESULT_IGNORED)) {
 					g_ptr_array_add (sorted_symbols, (gpointer)sym);
 				}
@@ -1419,8 +1424,8 @@ rspamd_task_log_variable (struct rspamd_task *task,
 		}
 		break;
 	case RSPAMD_LOG_USER:
-		if (task->user) {
-			var.begin = task->user;
+		if (task->auth_user) {
+			var.begin = task->auth_user;
 			var.len = strlen (var.begin);
 		}
 		else {
@@ -1668,12 +1673,12 @@ rspamd_task_get_required_score (struct rspamd_task *task, struct rspamd_scan_res
 	}
 
 	for (i = m->nactions - 1; i >= 0; i --) {
-		struct rspamd_action_result *action_lim = &m->actions_limits[i];
+		struct rspamd_action_config *action_lim = &m->actions_config[i];
 
 
 		if (!isnan (action_lim->cur_limit) &&
 				!(action_lim->action->flags & (RSPAMD_ACTION_NO_THRESHOLD|RSPAMD_ACTION_HAM))) {
-			return m->actions_limits[i].cur_limit;
+			return m->actions_config[i].cur_limit;
 		}
 	}
 
@@ -1906,7 +1911,7 @@ rspamd_task_timeout (EV_P_ ev_timer *w, int revents)
 
 		ev_timer_again (EV_A_ w);
 		task->processed_stages |= RSPAMD_TASK_STAGE_FILTERS;
-		rspamd_session_cleanup (task->s);
+		rspamd_session_cleanup (task->s, true);
 		rspamd_task_process (task, RSPAMD_TASK_PROCESS_ALL);
 		rspamd_session_pending (task->s);
 	}
@@ -1935,7 +1940,7 @@ rspamd_task_timeout (EV_P_ ev_timer *w, int revents)
 
 		ev_timer_stop (EV_A_ w);
 		task->processed_stages |= RSPAMD_TASK_STAGE_DONE;
-		rspamd_session_cleanup (task->s);
+		rspamd_session_cleanup (task->s, true);
 		rspamd_task_process (task, RSPAMD_TASK_PROCESS_ALL);
 		rspamd_session_pending (task->s);
 	}
@@ -1981,4 +1986,4 @@ rspamd_worker_guard_handler (EV_P_ ev_io *w, int revents)
 			return;
 		}
 	}
-}
\ No newline at end of file
+}
