@@ -32,7 +32,7 @@ extern "C" {
 enum rspamd_command {
 	CMD_SKIP = 0,
 	CMD_PING,
-	CMD_CHECK_SPAMC, /* Legacy spamasassin format */
+	CMD_CHECK_SPAMC, /* Legacy spamassassin format */
 	CMD_CHECK_RSPAMC, /* Legacy rspamc format (like SA one) */
 	CMD_CHECK, /* Legacy check - metric json reply */
 	CMD_CHECK_V2, /* Modern check - symbols in json reply  */
@@ -173,7 +173,7 @@ struct rspamd_task {
 	rspamd_inet_addr_t *from_addr;                    /**< from addr for a task							*/
 	rspamd_inet_addr_t *client_addr;                /**< address of connected socket					*/
 	gchar *deliver_to;                                /**< address to deliver								*/
-	gchar *user;                                    /**< user to deliver								*/
+	gchar *auth_user;                                /**< SMTP authenticated user								*/
 	const gchar *hostname;                            /**< hostname reported by MTA						*/
 	khash_t(rspamd_req_headers_hash) *request_headers; /**< HTTP headers in a request						*/
 	struct rspamd_task_data_storage msg;            /**< message buffer									*/
