@@ -21,6 +21,10 @@
 
 #define RSPAMD_SHINGLE_SIZE 32
 
+#ifdef  __cplusplus
+extern "C" {
+#endif
+
 struct rspamd_shingle {
 	guint64 hashes[RSPAMD_SHINGLE_SIZE];
 };
@@ -39,40 +43,40 @@ enum rspamd_shingle_alg {
  * @return shingle value
  */
 typedef guint64 (*rspamd_shingles_filter) (guint64 *input, gsize count,
-		gint shno, const guchar *key, gpointer ud);
+										   gint shno, const guchar *key, gpointer ud);
 
 /**
  * Generate shingles from the input of fixed size strings using lemmatizer
  * if needed
  * @param input array of `rspamd_fstring_t`
  * @param key secret key used to generate shingles
- * @param pool pool to allocate shigles array
+ * @param pool pool to allocate shingles array
  * @param filter hashes filtering function
  * @param filterd opaque data for filtering function
  * @return shingles array
  */
-struct rspamd_shingle* rspamd_shingles_from_text (GArray *input,
-		const guchar key[16],
-		rspamd_mempool_t *pool,
-		rspamd_shingles_filter filter,
-		gpointer filterd,
-		enum rspamd_shingle_alg alg);
+struct rspamd_shingle *rspamd_shingles_from_text (GArray *input,
+												  const guchar key[16],
+												  rspamd_mempool_t *pool,
+												  rspamd_shingles_filter filter,
+												  gpointer filterd,
+												  enum rspamd_shingle_alg alg);
 
 /**
  * Generate shingles from the DCT matrix of an image
  * @param dct discrete cosine transfor matrix (must be 64x64)
  * @param key secret key used to generate shingles
- * @param pool pool to allocate shigles array
+ * @param pool pool to allocate shingles array
  * @param filter hashes filtering function
  * @param filterd opaque data for filtering function
  * @return shingles array
  */
-struct rspamd_shingle* rspamd_shingles_from_image (guchar *dct,
-		const guchar key[16],
-		rspamd_mempool_t *pool,
-		rspamd_shingles_filter filter,
-		gpointer filterd,
-		enum rspamd_shingle_alg alg);
+struct rspamd_shingle *rspamd_shingles_from_image (guchar *dct,
+												   const guchar key[16],
+												   rspamd_mempool_t *pool,
+												   rspamd_shingles_filter filter,
+												   gpointer filterd,
+												   enum rspamd_shingle_alg alg);
 
 /**
  * Compares two shingles and return result as a floating point value - 1.0
@@ -82,12 +86,16 @@ struct rspamd_shingle* rspamd_shingles_from_image (guchar *dct,
  * @return
  */
 gdouble rspamd_shingles_compare (const struct rspamd_shingle *a,
-		const struct rspamd_shingle *b);
+								 const struct rspamd_shingle *b);
 
 /**
  * Default filtering function
  */
 guint64 rspamd_shingles_default_filter (guint64 *input, gsize count,
-		gint shno, const guchar *key, gpointer ud);
+										gint shno, const guchar *key, gpointer ud);
+
+#ifdef  __cplusplus
+}
+#endif
 
 #endif /* SHINGLES_H_ */
