@@ -169,7 +169,7 @@ lua_dns_resolver_callback (struct rdns_reply *reply, gpointer arg)
 	 * 3 - entries | nil
 	 * 4 - error | nil
 	 * 5 - user_str
-	 * 6 - reply->authenticated
+	 * 6 - reply->flags & RDNS_AUTH
 	 * 7 - server
 	 */
 	if (reply->code != RDNS_RC_NOERROR) {
@@ -183,7 +183,7 @@ lua_dns_resolver_callback (struct rdns_reply *reply, gpointer arg)
 		lua_pushnil (L);
 	}
 
-	lua_pushboolean (L, reply->authenticated);
+	lua_pushboolean (L, reply->flags & RDNS_AUTH);
 
 	const gchar *servname = rdns_request_get_server (reply->request);
 
@@ -504,16 +504,18 @@ lua_dns_resolver_resolve_common (lua_State *L,
 	}
 
 	return 1;
+
 err:
+	/* Callback is not called in this case */
+	if (cbdata->cbref != -1) {
+		luaL_unref (L, LUA_REGISTRYINDEX, cbdata->cbref);
+	}
+
 	if (!pool) {
 		/* Free resources */
 		g_free (cbdata->to_resolve);
 		g_free (cbdata->user_str);
-	}
-
-	/* Callback is not called in this case */
-	if (cbdata->cbref != -1) {
-		luaL_unref (L, LUA_REGISTRYINDEX, cbdata->cbref);
+		g_free (cbdata);
 	}
 
 	lua_pushnil (L);
@@ -530,7 +532,7 @@ err:
  * * `mempool` - pool memory pool for storing intermediate data
  * * `name` - host name to resolve
  * * `callback` - callback callback function to be called upon name resolution is finished; must be of type `function (resolver, to_resolve, results, err)`
- * * `forced` - true if needed to override notmal limit for DNS requests
+ * * `forced` - true if needed to override normal limit for DNS requests
  * @return {boolean} `true` if DNS request has been scheduled
  */
 static int
@@ -560,7 +562,7 @@ lua_dns_resolver_resolve_a (lua_State *L)
  * * `mempool` - pool memory pool for storing intermediate data
  * * `name` - host name to resolve
  * * `callback` - callback callback function to be called upon name resolution is finished; must be of type `function (resolver, to_resolve, results, err)`
- * * `forced` - true if needed to override notmal limit for DNS requests
+ * * `forced` - true if needed to override normal limit for DNS requests
  * @return {boolean} `true` if DNS request has been scheduled
  */
 static int
@@ -590,7 +592,7 @@ lua_dns_resolver_resolve_ptr (lua_State *L)
  * * `mempool` - pool memory pool for storing intermediate data
  * * `name` - host name to resolve
  * * `callback` - callback callback function to be called upon name resolution is finished; must be of type `function (resolver, to_resolve, results, err)`
- * * `forced` - true if needed to override notmal limit for DNS requests
+ * * `forced` - true if needed to override normal limit for DNS requests
  * @return {boolean} `true` if DNS request has been scheduled
  */
 static int
@@ -620,7 +622,7 @@ lua_dns_resolver_resolve_txt (lua_State *L)
  * * `mempool` - pool memory pool for storing intermediate data
  * * `name` - host name to resolve
  * * `callback` - callback callback function to be called upon name resolution is finished; must be of type `function (resolver, to_resolve, results, err)`
- * * `forced` - true if needed to override notmal limit for DNS requests
+ * * `forced` - true if needed to override normal limit for DNS requests
  * @return {boolean} `true` if DNS request has been scheduled
  */
 static int
@@ -650,7 +652,7 @@ lua_dns_resolver_resolve_mx (lua_State *L)
  * * `mempool` - pool memory pool for storing intermediate data
  * * `name` - host name to resolve
  * * `callback` - callback callback function to be called upon name resolution is finished; must be of type `function (resolver, to_resolve, results, err)`
- * * `forced` - true if needed to override notmal limit for DNS requests
+ * * `forced` - true if needed to override normal limit for DNS requests
  * @return {boolean} `true` if DNS request has been scheduled
  */
 static int
