@@ -134,6 +134,10 @@ lua_http_fin (gpointer arg)
 		g_free (cbd->auth);
 	}
 
+	if (cbd->host) {
+		g_free (cbd->host);
+	}
+
 	if (cbd->local_kp) {
 		rspamd_keypair_unref (cbd->local_kp);
 	}
@@ -388,14 +392,20 @@ static gboolean
 lua_http_make_connection (struct lua_http_cbdata *cbd)
 {
 	rspamd_inet_address_set_port (cbd->addr, cbd->msg->port);
+	unsigned http_opts = RSPAMD_HTTP_CLIENT_SIMPLE;
+
+	if (cbd->msg->flags & RSPAMD_HTTP_FLAG_WANT_SSL) {
+		http_opts |= RSPAMD_HTTP_CLIENT_SSL;
+	}
 
 	if (cbd->flags & RSPAMD_LUA_HTTP_FLAG_KEEP_ALIVE) {
 		cbd->fd = -1; /* FD is owned by keepalive connection */
-		cbd->conn = rspamd_http_connection_new_keepalive (
+		cbd->conn = rspamd_http_connection_new_client_keepalive(
 				NULL, /* Default context */
 				NULL,
 				lua_http_error_handler,
 				lua_http_finish_handler,
+				http_opts,
 				cbd->addr,
 				cbd->host);
 	}
@@ -406,7 +416,7 @@ lua_http_make_connection (struct lua_http_cbdata *cbd)
 				NULL,
 				lua_http_error_handler,
 				lua_http_finish_handler,
-				RSPAMD_HTTP_CLIENT_SIMPLE,
+				http_opts,
 				cbd->addr);
 	}
 
@@ -433,23 +443,30 @@ lua_http_make_connection (struct lua_http_cbdata *cbd)
 		}
 
 		if (cbd->session) {
-			rspamd_session_add_event (cbd->session,
-					(event_finalizer_t) lua_http_fin, cbd,
-					M);
+			if (cbd->item) {
+				rspamd_session_add_event_full (cbd->session,
+						(event_finalizer_t) lua_http_fin, cbd,
+						M, rspamd_symcache_item_name (cbd->item));
+			}
+			else {
+				rspamd_session_add_event (cbd->session,
+						(event_finalizer_t) lua_http_fin, cbd,
+						M);
+			}
 			cbd->flags |= RSPAMD_LUA_HTTP_FLAG_RESOLVED;
 		}
 
 		if (cbd->task) {
 			cbd->conn->log_tag = cbd->task->task_pool->tag.uid;
+
+			if (cbd->item) {
+				rspamd_symcache_item_async_inc (cbd->task, cbd->item, M);
+			}
 		}
 		else if (cbd->cfg) {
 			cbd->conn->log_tag = cbd->cfg->cfg_pool->tag.uid;
 		}
 
-		if (cbd->item) {
-			rspamd_symcache_item_async_inc (cbd->task, cbd->item, M);
-		}
-
 		struct rspamd_http_message *msg = cbd->msg;
 
 		/* Message is now owned by a connection object */
@@ -652,10 +669,13 @@ lua_http_request (lua_State *L)
 
 		if (lua_type (L, -1) == LUA_TUSERDATA) {
 			task = lua_check_task (L, -1);
-			ev_base = task->event_loop;
-			resolver = task->resolver;
-			session = task->s;
-			cfg = task->cfg;
+
+			if (task) {
+				ev_base = task->event_loop;
+				resolver = task->resolver;
+				session = task->s;
+				cfg = task->cfg;
+			}
 		}
 		lua_pop (L, 1);
 
@@ -823,7 +843,10 @@ lua_http_request (lua_State *L)
 				}
 				else {
 					t = lua_check_text (L, -1);
-					body = rspamd_fstring_append (body, t->start, t->len);
+
+					if (t) {
+						body = rspamd_fstring_append(body, t->start, t->len);
+					}
 				}
 
 				lua_pop (L, 1);
@@ -963,6 +986,9 @@ lua_http_request (lua_State *L)
 		if (body) {
 			rspamd_fstring_free (body);
 		}
+		if (local_kp) {
+			rspamd_keypair_unref (local_kp);
+		}
 
 		return 1;
 	}
@@ -972,12 +998,24 @@ lua_http_request (lua_State *L)
 		if (body) {
 			rspamd_fstring_free (body);
 		}
+		if (local_kp) {
+			rspamd_keypair_unref (local_kp);
+		}
 
 		return luaL_error (L,
 				"Bad params to rspamd_http:request(): either task or config should be set");
 	}
 
 	if (ev_base == NULL) {
+		g_free (auth);
+		rspamd_http_message_unref (msg);
+		if (body) {
+			rspamd_fstring_free (body);
+		}
+		if (local_kp) {
+			rspamd_keypair_unref (local_kp);
+		}
+
 		return luaL_error (L,
 				"Bad params to rspamd_http:request(): ev_base isn't passed");
 	}
@@ -1008,9 +1046,6 @@ lua_http_request (lua_State *L)
 		cbd->item = rspamd_symcache_get_cur_item (task);
 	}
 
-	if (msg->host) {
-		cbd->host = msg->host->str;
-	}
 
 	if (body) {
 		if (gzip) {
@@ -1026,8 +1061,42 @@ lua_http_request (lua_State *L)
 		cbd->session = session;
 	}
 
-	if (rspamd_parse_inet_address (&cbd->addr,
-			msg->host->str, msg->host->len, RSPAMD_INET_ADDRESS_PARSE_DEFAULT)) {
+	bool numeric_ip = false;
+
+	/* Check if we can skip resolving */
+	gsize hostlen = 0;
+	const gchar *host = rspamd_http_message_get_http_host (msg, &hostlen);
+
+	if (host) {
+		cbd->host = g_malloc (hostlen + 1);
+		rspamd_strlcpy (cbd->host, host, hostlen + 1);
+
+		if (cbd->flags & RSPAMD_LUA_HTTP_FLAG_KEEP_ALIVE) {
+			const rspamd_inet_addr_t *ka_addr = rspamd_http_context_has_keepalive(NULL,
+					cbd->host,
+					msg->port,
+					msg->flags & RSPAMD_HTTP_FLAG_WANT_SSL);
+
+			if (ka_addr) {
+				cbd->addr = rspamd_inet_address_copy(ka_addr, NULL);
+				numeric_ip = true;
+			}
+		}
+
+		if (!cbd->addr) {
+			/* We use msg->host here, not cbd->host ! */
+			if (rspamd_parse_inet_address (&cbd->addr,
+					msg->host->str, msg->host->len,
+					RSPAMD_INET_ADDRESS_PARSE_DEFAULT)) {
+				numeric_ip = true;
+			}
+		}
+	}
+	else {
+		cbd->host = NULL;
+	}
+
+	if (numeric_ip) {
 		/* Host is numeric IP, no need to resolve */
 		gboolean ret;
 
