@@ -89,7 +89,6 @@ static const struct luaL_reg redislib_m[] = {
 #define REDIS_RELEASE REF_RELEASE
 #endif
 
-#ifdef WITH_HIREDIS
 struct lua_redis_request_specific_userdata;
 /**
  * Struct for userdata representation
@@ -189,7 +188,7 @@ lua_redis_dtor (struct lua_redis_ctx *ctx)
 	struct redisAsyncContext *ac;
 
 	ud = &ctx->async;
-	msg_debug_lua_redis ("desctructing %p", ctx);
+	msg_debug_lua_redis ("destructing %p", ctx);
 
 	if (ud->ctx) {
 
@@ -612,75 +611,86 @@ lua_redis_callback_sync (redisAsyncContext *ac, gpointer r, gpointer priv)
 		ev_timer_stop (ud->event_loop, &sp_ud->timeout_ev);
 	}
 
-	msg_debug_lua_redis ("got reply from redis: %p for query %p", ac, sp_ud);
+	if (!(sp_ud->flags & LUA_REDIS_SPECIFIC_FINISHED)) {
+		msg_debug_lua_redis ("got reply from redis: %p for query %p", ac, sp_ud);
 
-	struct lua_redis_result *result = g_malloc0 (sizeof *result);
+		struct lua_redis_result *result = g_malloc0 (sizeof *result);
 
-	/* If session is finished, we cannot call lua callbacks */
-	if (ac->err == 0) {
-		if (r != NULL) {
-			if (reply->type != REDIS_REPLY_ERROR) {
-				result->is_error = FALSE;
-				lua_redis_push_reply (L, reply, ctx->flags & LUA_REDIS_TEXTDATA);
+		if (ac->err == 0) {
+			if (r != NULL) {
+				if (reply->type != REDIS_REPLY_ERROR) {
+					result->is_error = FALSE;
+					lua_redis_push_reply (L, reply, ctx->flags & LUA_REDIS_TEXTDATA);
+				}
+				else {
+					result->is_error = TRUE;
+					lua_pushstring (L, reply->str);
+				}
 			}
 			else {
 				result->is_error = TRUE;
-				lua_pushstring (L, reply->str);
+				lua_pushliteral (L, "received no data from server");
 			}
 		}
 		else {
 			result->is_error = TRUE;
-			lua_pushliteral (L, "received no data from server");
+			if (ac->err == REDIS_ERR_IO) {
+				lua_pushstring (L, strerror (errno));
+			}
+			else {
+				lua_pushstring (L, ac->errstr);
+			}
 		}
-	}
-	else {
-		result->is_error = TRUE;
-		if (ac->err == REDIS_ERR_IO) {
-			lua_pushstring (L, strerror (errno));
-		}
-		else {
-			lua_pushstring (L, ac->errstr);
+
+		/* if error happened, we should terminate the connection,
+		   and release it */
+
+		if (result->is_error && sp_ud->c->ctx) {
+			ac = sp_ud->c->ctx;
+			/* Set to NULL to avoid double free in dtor */
+			sp_ud->c->ctx = NULL;
+			ctx->flags |= LUA_REDIS_TERMINATED;
+
+			/*
+			 * This will call all callbacks pending so the entire context
+			 * will be destructed
+			 */
+			rspamd_redis_pool_release_connection (sp_ud->c->pool, ac,
+					RSPAMD_REDIS_RELEASE_FATAL);
 		}
+
+		result->result_ref = luaL_ref (L, LUA_REGISTRYINDEX);
+		result->s = ud->s;
+		result->item = ud->item;
+		result->task = ud->task;
+		result->sp_ud = sp_ud;
+
+		g_queue_push_tail (ctx->replies, result);
+
 	}
 
-	/* if error happened, we should terminate the connection,
-	   and release it */
-
-	if (result->is_error && sp_ud->c->ctx) {
-		ac = sp_ud->c->ctx;
-		/* Set to NULL to avoid double free in dtor */
-		sp_ud->c->ctx = NULL;
-		ctx->flags |= LUA_REDIS_TERMINATED;
-
-		/*
-		 * This will call all callbacks pending so the entire context
-		 * will be destructed
-		 */
-		rspamd_redis_pool_release_connection (sp_ud->c->pool, ac,
-				RSPAMD_REDIS_RELEASE_FATAL);
-	}
-
-	result->result_ref = luaL_ref (L, LUA_REGISTRYINDEX);
-	result->s = ud->s;
-	result->item = ud->item;
-	result->task = ud->task;
-	result->sp_ud = sp_ud;
-
-	g_queue_push_tail (ctx->replies, result);
-
 	ctx->cmds_pending --;
 
 	if (ctx->cmds_pending == 0) {
 		if (ctx->thread) {
-			/* somebody yielded and waits for results */
-			thread = ctx->thread;
-			ctx->thread = NULL;
+			if (!(sp_ud->flags & LUA_REDIS_SPECIFIC_FINISHED)) {
+				/* somebody yielded and waits for results */
+				thread = ctx->thread;
+				ctx->thread = NULL;
 
-			results = lua_redis_push_results (ctx, thread->lua_state);
-			lua_thread_resume (thread, results);
-			lua_redis_cleanup_events (ctx);
+				results = lua_redis_push_results(ctx, thread->lua_state);
+				lua_thread_resume (thread, results);
+				lua_redis_cleanup_events(ctx);
+			}
+			else {
+				/* We cannot resume the thread as the associated task has gone */
+				lua_thread_pool_terminate_entry_full (ud->cfg->lua_thread_pool,
+						ctx->thread, G_STRLOC, true);
+				ctx->thread = NULL;
+			}
 		}
 	}
+
 }
 
 static void
@@ -692,6 +702,10 @@ lua_redis_timeout_sync (EV_P_ ev_timer *w, int revents)
 	struct lua_redis_userdata *ud;
 	redisAsyncContext *ac;
 
+	if (sp_ud->flags & LUA_REDIS_SPECIFIC_FINISHED) {
+		return;
+	}
+
 	ud = sp_ud->c;
 	ctx = sp_ud->ctx;
 	msg_debug_lua_redis ("timeout while querying redis server: %p, redis: %p", sp_ud,
@@ -1333,6 +1347,8 @@ lua_redis_make_request_sync (lua_State *L)
 			rspamd_inet_address_free (ip);
 		}
 		msg_err ("bad arguments for redis request");
+		lua_redis_free_args (args, arglens, nargs);
+
 		lua_pushboolean (L, FALSE);
 	}
 
@@ -1595,67 +1611,6 @@ lua_redis_exec (lua_State *L)
 		}
 	}
 }
-#else
-static int
-lua_redis_make_request (lua_State *L)
-{
-	msg_warn ("rspamd is compiled with no redis support");
-
-	lua_pushboolean (L, FALSE);
-
-	return 1;
-}
-static int
-lua_redis_make_request_sync (lua_State *L)
-{
-	msg_warn ("rspamd is compiled with no redis support");
-
-	lua_pushboolean (L, FALSE);
-
-	return 1;
-}
-static int
-lua_redis_connect (lua_State *L)
-{
-	msg_warn ("rspamd is compiled with no redis support");
-
-	lua_pushboolean (L, FALSE);
-
-	return 1;
-}
-static int
-lua_redis_connect_sync (lua_State *L)
-{
-	msg_warn ("rspamd is compiled with no redis support");
-
-	lua_pushboolean (L, FALSE);
-
-	return 1;
-}
-static int
-lua_redis_add_cmd (lua_State *L)
-{
-	msg_warn ("rspamd is compiled with no redis support");
-
-	lua_pushboolean (L, FALSE);
-
-	return 1;
-}
-static int
-lua_redis_exec (lua_State *L)
-{
-	msg_warn ("rspamd is compiled with no redis support");
-
-	lua_pushboolean (L, FALSE);
-
-	return 1;
-}
-static int
-lua_redis_gc (lua_State *L)
-{
-	return 0;
-}
-#endif
 
 static gint
 lua_load_redis (lua_State * L)
