@@ -100,7 +100,7 @@ LUA_FUNCTION_DEF (tcp, request);
  * @function rspamd_tcp.connect_sync()
  *
  * Creates pseudo-synchronous TCP connection.
- * Each method of the connection requiring IO, becames a yielding point,
+ * Each method of the connection requiring IO, becomes a yielding point,
  * i.e. current thread Lua thread is get suspended and resumes as soon as IO is done
  *
  * This class represents low-level API, using of "lua_tcp_sync" module is recommended.
@@ -785,6 +785,7 @@ lua_tcp_write_helper (struct lua_tcp_cbdata *cbd)
 	struct iovec *start;
 	guint niov, i;
 	gint flags = 0;
+	bool allocated_iov = false;
 	gsize remain;
 	gssize r;
 	struct iovec *cur_iov;
@@ -811,6 +812,7 @@ lua_tcp_write_helper (struct lua_tcp_cbdata *cbd)
 	}
 	else {
 		cur_iov = g_malloc0 (niov * sizeof (struct iovec));
+		allocated_iov = true;
 	}
 
 	memcpy (cur_iov, wh->iov, niov * sizeof (struct iovec));
@@ -848,7 +850,7 @@ lua_tcp_write_helper (struct lua_tcp_cbdata *cbd)
 		r = sendmsg (cbd->fd, &msg, flags);
 	}
 
-	if (niov >= 1024) {
+	if (allocated_iov) {
 		g_free (cur_iov);
 	}
 
@@ -1229,7 +1231,13 @@ lua_tcp_register_event (struct lua_tcp_cbdata *cbd)
 	if (cbd->session) {
 		event_finalizer_t fin = IS_SYNC (cbd) ? lua_tcp_void_finalyser : lua_tcp_fin;
 
-		cbd->async_ev = rspamd_session_add_event (cbd->session, fin, cbd, M);
+		if (cbd->item) {
+			cbd->async_ev = rspamd_session_add_event_full (cbd->session, fin, cbd, M,
+					rspamd_symcache_item_name (cbd->item));
+		}
+		else {
+			cbd->async_ev = rspamd_session_add_event (cbd->session, fin, cbd, M);
+		}
 
 		if (!cbd->async_ev) {
 			return FALSE;
@@ -1242,7 +1250,7 @@ lua_tcp_register_event (struct lua_tcp_cbdata *cbd)
 static void
 lua_tcp_register_watcher (struct lua_tcp_cbdata *cbd)
 {
-	if (cbd->item) {
+	if (cbd->item && cbd->task) {
 		rspamd_symcache_item_async_inc (cbd->task, cbd->item, M);
 	}
 }
@@ -1519,6 +1527,8 @@ lua_tcp_request (lua_State *L)
 				event_loop = *(struct ev_loop **)lua_touserdata (L, -1);
 			}
 			else {
+				g_free (cbd);
+
 				return luaL_error (L, "event loop is required");
 			}
 			lua_pop (L, 1);
@@ -1690,6 +1700,9 @@ lua_tcp_request (lua_State *L)
 	}
 
 	if (resolver == NULL && cfg == NULL && task == NULL) {
+		g_free (cbd);
+		g_free (iov);
+
 		return luaL_error (L, "tcp request has bad params: one of "
 						"{resolver,task,config} should be set");
 	}
