@@ -236,8 +236,14 @@ static gboolean
 lua_udp_maybe_register_event (struct lua_udp_cbdata *cbd)
 {
 	if (cbd->s && !cbd->async_ev) {
-		cbd->async_ev = rspamd_session_add_event (cbd->s, lua_udp_cbd_fin,
-				cbd, M);
+		if (cbd->item) {
+			cbd->async_ev = rspamd_session_add_event_full (cbd->s, lua_udp_cbd_fin,
+					cbd, M, rspamd_symcache_item_name (cbd->item));
+		}
+		else {
+			cbd->async_ev = rspamd_session_add_event (cbd->s, lua_udp_cbd_fin,
+					cbd, M);
+		}
 
 		if (!cbd->async_ev) {
 			return FALSE;
@@ -391,7 +397,7 @@ lua_udp_sendto (lua_State *L) {
 				return luaL_error (L, "invalid host class");
 			}
 
-			addr = rspamd_inet_address_copy (lip->addr);
+			addr = rspamd_inet_address_copy(lip->addr, NULL);
 
 			if (port != 0) {
 				rspamd_inet_address_set_port (addr, port);
